# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm
from niapy.util.distances import euclidean
import niapy.util.repair as repair

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['FireworksAlgorithm', 'EnhancedFireworksAlgorithm', 'DynamicFireworksAlgorithm',
           'DynamicFireworksAlgorithmGauss', 'BareBonesFireworksAlgorithm']


class BareBonesFireworksAlgorithm(Algorithm):
    r"""Implementation of Bare Bones Fireworks Algorithm.

    Algorithm:
        Bare Bones Fireworks Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.sciencedirect.com/science/article/pii/S1568494617306609

    Reference paper:
        Junzhi Li, Ying Tan, The bare bones fireworks algorithm: A minimalist global optimizer, Applied Soft Computing, Volume 62, 2018, Pages 454-462, ISSN 1568-4946, https://doi.org/10.1016/j.asoc.2017.10.046.

    Attributes:
        Name (List[str]): List of strings representing algorithm names
        num_sparks (int): Number of sparks
        amplification_coefficient (float): amplification coefficient
        reduction_coefficient (float): reduction coefficient

    """

    Name = ['BareBonesFireworksAlgorithm', 'BBFWA']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Junzhi Li, Ying Tan, The bare bones fireworks algorithm: A minimalist global optimizer, Applied Soft Computing, Volume 62, 2018, Pages 454-462, ISSN 1568-4946, https://doi.org/10.1016/j.asoc.2017.10.046."""

    def __init__(self, num_sparks=10, amplification_coefficient=1.5, reduction_coefficient=0.5, *args, **kwargs):
        r"""Initialize BareBonesFireworksAlgorithm.

        Args:
            num_sparks (int): Number of sparks :math:`\in[1, \infty)`.
            amplification_coefficient (float): Amplification coefficient :math:`\in [1, \infty)`.
            reduction_coefficient (float): Reduction coefficient :math:`\in (0, 1)`.

        """
        kwargs.pop('population_size', None)
        super().__init__(1, *args, **kwargs)
        self.num_sparks = num_sparks
        self.amplification_coefficient = amplification_coefficient
        self.reduction_coefficient = reduction_coefficient

    def set_parameters(self, num_sparks=10, amplification_coefficient=1.5, reduction_coefficient=0.5, **kwargs):
        r"""Set the arguments of an algorithm.

        Args:
            num_sparks (int): Number of sparks :math:`\in [1, \infty)`.
            amplification_coefficient (float): Amplification coefficient :math:`\in [1, \infty)`.
            reduction_coefficient (float): Reduction coefficient :math:`\in (0, 1)`.

        """
        kwargs.pop('population_size', None)
        super().set_parameters(population_size=1, **kwargs)
        self.num_sparks = num_sparks
        self.amplification_coefficient = amplification_coefficient
        self.reduction_coefficient = reduction_coefficient

    def init_population(self, task):
        r"""Initialize starting population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, float, Dict[str, Any]]:
                1. Initial solution.
                2. Initial solution function/fitness value.
                3. Additional arguments:
                    * A (numpy.ndarray): Starting amplitude or search range.

        """
        x, x_fit, d = super().init_population(task)
        d.update({'amplitude': task.range})
        return x, x_fit, d

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Bare Bones Fireworks Algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current solution.
            population_fitness (float): Current solution fitness/function value.
            best_x (numpy.ndarray): Current best solution.
            best_fitness (float): Current best solution fitness/function value.
            params (Dict[str, Any]): Additional parameters.

        Returns:
            Tuple[numpy.ndarray, float, numpy.ndarray, float, Dict[str, Any]]:
                1. New solution.
                2. New solution fitness/function value.
                3. New global best solution.
                4. New global best solutions fitness/objective value.
                5. Additional arguments:
                    * amplitude (numpy.ndarray): Search range.

        """
        amplitude = params.pop('amplitude')

        sparks = self.uniform(population - amplitude, population + amplitude, (self.num_sparks, task.dimension))
        sparks = np.apply_along_axis(task.repair, 1, sparks, self.rng)
        sparks_fitness = np.apply_along_axis(task.eval, 1, sparks)
        best_index = np.argmin(sparks_fitness)
        if sparks_fitness[best_index] < population_fitness:
            population = sparks[best_index]
            population_fitness = sparks_fitness[best_index]
            amplitude = self.amplification_coefficient * amplitude
        else:
            amplitude = self.reduction_coefficient * amplitude
        return population, population_fitness, population.copy(), population_fitness, {'amplitude': amplitude}


class FireworksAlgorithm(Algorithm):
    r"""Implementation of fireworks algorithm.

    Algorithm:
        Fireworks Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.springer.com/gp/book/9783662463529

    Reference paper:
        Tan, Ying. "Fireworks algorithm." Heidelberg, Germany: Springer 10 (2015): 978-3

    Attributes:
        Name (List[str]): List of strings representing algorithm names.

    """

    Name = ['FireworksAlgorithm', 'FWA']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Tan, Ying. "Fireworks algorithm." Heidelberg, Germany: Springer 10 (2015): 978-3."""

    def __init__(self, population_size=5, num_sparks=50, a=0.04, b=0.8, max_amplitude=40, num_gaussian=5, *args, **kwargs):
        """Initialize FWA.

        Args:
            population_size (int): Number of Fireworks
            num_sparks (int): Number of sparks
            a (float): Limitation of sparks
            b (float): Limitation of sparks
            max_amplitude (float): Initial amplitude.
            num_gaussian (int): Number of sparks to apply gaussian mutation to.

        """
        super().__init__(population_size, *args, **kwargs)
        self.num_sparks = num_sparks
        self.a = a
        self.b = b
        self.max_amplitude = max_amplitude
        self.num_gaussian = num_gaussian
        self.epsilon = np.finfo(float).eps

    def set_parameters(self, population_size=5, num_sparks=50, a=0.04, b=0.8, max_amplitude=40, num_gaussian=5,
                       **kwargs):
        r"""Set the arguments of an algorithm.

        Args:
            population_size (int): Number of Fireworks
            num_sparks (int): Number of sparks
            a (float): Limitation of sparks
            b (float): Limitation of sparks
            max_amplitude (float): Initial amplitude.
            num_gaussian (int): Number of sparks to apply gaussian mutation to.

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.num_sparks = num_sparks
        self.a = a
        self.b = b
        self.max_amplitude = max_amplitude
        self.num_gaussian = num_gaussian
        self.epsilon = np.finfo(float).eps

    def sparks_num(self, population_fitness):
        r"""Calculate number of sparks.

        Args:
            population_fitness (numpy.ndarray): Population fitness values.

        Returns:
            numpy.ndarray: Number of sparks that for all fireworks.

        """
        worst_fitness = np.amax(population_fitness)
        sparks_num = self.num_sparks * (worst_fitness - population_fitness + self.epsilon)
        sparks_num /= np.sum(worst_fitness - population_fitness) + self.epsilon

        cond = [sparks_num < self.a * self.num_sparks, (sparks_num > self.b * self.num_sparks) * (self.a < self.b < 1)]
        choices = [round(self.a * self.num_sparks), round(self.b * self.num_sparks)]
        return np.select(cond, choices, default=np.round(sparks_num)).astype(int)

    def explosion_amplitudes(self, population_fitness, task=None):
        r"""Calculate explosion amplitude.

        Args:
            population_fitness (numpy.ndarray): Population fitness values.
            task (Optional[Task]): Optimization task (Unused in this version of the algorithm).

        Returns:
            numpy.ndarray: Explosion amplitude of sparks.

        """
        best_fitness = np.amin(population_fitness)
        amplitudes = self.max_amplitude * (population_fitness - best_fitness + self.epsilon)
        amplitudes /= np.sum(population_fitness - best_fitness) + self.epsilon
        return amplitudes

    def explosion_spark(self, x, amplitude, task):
        r"""Explode a spark.

        Args:
            x (numpy.ndarray): Individuals creating spark.
            amplitude (float): Amplitude of spark.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Sparks exploded in with specified amplitude.

        """
        z = self.rng.choice(task.dimension, self.rng.integers(task.dimension), replace=False)
        x[z] = x[z] + amplitude * self.uniform(-1, 1)
        return self.mapping(x, task)

    def gaussian_spark(self, x, task, best_x=None):
        r"""Create gaussian spark.

        Args:
            x (numpy.ndarray): Individual creating a spark.
            task (Task): Optimization task.
            best_x (numpy.ndarray): Current best individual. Unused in this version of the algorithm.

        Returns:
            numpy.ndarray: Spark exploded based on gaussian amplitude.

        """
        z = self.rng.choice(task.dimension, self.rng.integers(task.dimension), replace=False)
        x[z] = x[z] * self.normal(1, 1)
        return self.mapping(x, task)

    def mapping(self, x, task):
        r"""Fix value to bounds.

        Args:
            x (numpy.ndarray): Individual to fix.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Individual in search range.

        """
        return repair.reflect(x, task.lower, task.upper)

    def selection(self, population, population_fitness, sparks, task):
        r"""Generate new generation of individuals.

        Args:
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray[float]): Currents population fitness/function values.
            sparks (numpy.ndarray): New population.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], numpy.ndarray, float]:
                1. New population.
                2. New populations fitness/function values.
                3. New global best individual.
                4. New global best fitness.

        """
        sparks_fitness = np.apply_along_axis(task.eval, 1, sparks)
        best_index = np.argmin(sparks_fitness)
        best_x = sparks[best_index].copy()
        best_fitness = sparks_fitness[best_index]

        all_sparks = np.delete(sparks, best_index, axis=0)
        fitness = np.delete(sparks_fitness, best_index)

        distances = np.sum(euclidean(all_sparks[:, np.newaxis, :], all_sparks[np.newaxis, :, :]), axis=0)
        probabilities = distances / np.sum(distances)

        selected_indices = self.rng.choice(len(all_sparks), self.population_size - 1, replace=False, p=probabilities)

        population[0] = best_x
        population[1:] = all_sparks[selected_indices]
        population_fitness[0] = best_fitness
        population_fitness[1:] = fitness[selected_indices]
        return population, population_fitness, best_x, best_fitness

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Fireworks algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray[float]): Current populations function/fitness values.
            best_x (numpy.ndarray): Global best individual.
            best_fitness (float): Global best individuals fitness/function value.
            **params (Dict[str, Any)]: Additional arguments

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. Initialized population.
                2. Initialized populations function/fitness values.
                3. New global best solution.
                4. New global best solutions fitness/objective value.
                5. Additional arguments:
                    * Ah (numpy.ndarray): Initialized amplitudes.

        See Also:
            * :func:`FireworksAlgorithm.sparks_num`.
            * :func:`FireworksAlgorithm.explosion_amplitudes`
            * :func:`FireworksAlgorithm.explosion_spark`
            * :func:`FireworksAlgorithm.gaussian_spark`
            * :func:`FireworksAlgorithm.selection`

        """
        sparks_num = self.sparks_num(population_fitness)
        amplitudes = self.explosion_amplitudes(population_fitness, task=task)

        all_sparks = population.copy()
        for i in range(self.population_size):
            si = sparks_num[i]
            ai = amplitudes[i]

            sparks_i = np.empty((si, task.dimension))
            for s in range(si):
                sparks_i[s] = population[i]
                sparks_i[s] = self.explosion_spark(sparks_i[s], ai, task)
            all_sparks = np.concatenate((all_sparks, sparks_i), axis=0)

        gaussian_idx = self.rng.choice(len(all_sparks), self.num_gaussian, replace=False)
        gaussian_sparks = np.array(all_sparks[gaussian_idx])

        for i in range(self.num_gaussian):
            gaussian_sparks[i] = self.gaussian_spark(gaussian_sparks[i], task, best_x=best_x)

        all_sparks = np.concatenate((all_sparks, gaussian_sparks), axis=0)

        population, population_fitness, best_x, best_fitness = self.selection(population, population_fitness,
                                                                              all_sparks, task)

        return population, population_fitness, best_x, best_fitness, {}


class EnhancedFireworksAlgorithm(FireworksAlgorithm):
    r"""Implementation of enhanced fireworks algorithm.

    Algorithm:
        Enhanced Fireworks Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://ieeexplore.ieee.org/document/6557813/

    Reference paper:
        S. Zheng, A. Janecek and Y. Tan, "Enhanced Fireworks Algorithm," 2013 IEEE Congress on Evolutionary Computation, Cancun, 2013, pp. 2069-2077. doi: 10.1109/CEC.2013.6557813

    Attributes:
        Name (List[str]): List of strings representing algorithm names.
        amplitude_init (float): Initial amplitude of sparks.
        amplitude_final (float): Maximal amplitude of sparks.

    """

    Name = ['EnhancedFireworksAlgorithm', 'EFWA']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""S. Zheng, A. Janecek and Y. Tan, "Enhanced Fireworks Algorithm," 2013 IEEE Congress on Evolutionary Computation, Cancun, 2013, pp. 2069-2077. doi: 10.1109/CEC.2013.6557813"""

    def __init__(self, amplitude_init=0.2, amplitude_final=0.01, *args, **kwargs):
        """Initialize EFWA.

        Args:
            amplitude_init (float): Initial amplitude.
            amplitude_final (float): Final amplitude.

        See Also:
            * :func:`FireworksAlgorithm.__init__`

        """
        super().__init__(*args, **kwargs)
        self.amplitude_init = amplitude_init
        self.amplitude_final = amplitude_final

    def set_parameters(self, amplitude_init=0.2, amplitude_final=0.01, **kwargs):
        r"""Set EnhancedFireworksAlgorithm algorithms core parameters.

        Args:
            amplitude_init (float): Initial amplitude.
            amplitude_final (float): Final amplitude.

        See Also:
            * :func:`FireworksAlgorithm.set_parameters`

        """
        super().set_parameters(**kwargs)
        self.amplitude_init = amplitude_init
        self.amplitude_final = amplitude_final

    def explosion_amplitudes(self, population_fitness, task=None):
        r"""Calculate explosion amplitude.

        Args:
            population_fitness (numpy.ndarray):
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New amplitude.

        """
        amplitudes = super().explosion_amplitudes(population_fitness, task)
        a_min = self.amplitude_init - np.sqrt(task.evals * (2 * task.max_evals - task.evals)) * (self.amplitude_init - self.amplitude_final) / task.max_evals
        amplitudes[amplitudes < a_min] = a_min
        return amplitudes

    def mapping(self, x, task):
        r"""Fix value to bounds.

        Args:
            x (numpy.ndarray): Individual to fix.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Individual in search range.

        """
        return repair.rand(x, task.lower, task.upper, rng=self.rng)

    def explosion_spark(self, x, amplitude, task):
        r"""Explode a spark.

        Args:
            x (numpy.ndarray): Individuals creating spark.
            amplitude (float): Amplitude of spark.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: Sparks exploded in with specified amplitude.

        """
        z = self.rng.choice(task.dimension, self.rng.integers(task.dimension), replace=False)
        if isinstance(amplitude, np.ndarray):
            x[z] = x[z] + amplitude[z] * self.uniform(-1, 1, len(z))
        else:
            x[z] = x[z] + amplitude * self.uniform(-1, 1, len(z))
        return self.mapping(x, task)

    def gaussian_spark(self, x, task, best_x=None):
        r"""Create new individual.

        Args:
            x (numpy.ndarray):
            task (Task): Optimization task.
            best_x (numpy.ndarray): Current global best individual.

        Returns:
            numpy.ndarray: New individual generated by gaussian noise.

        """
        z = self.rng.choice(task.dimension, self.rng.integers(task.dimension), replace=False)
        e = self.standard_normal()
        x[z] = x[z] + (best_x[z] - x[z]) * e
        return self.mapping(x, task)

    def selection(self, population, population_fitness, sparks, task):
        r"""Generate new population.

        Args:
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray[float]): Current populations fitness/function values.
            sparks (numpy.ndarray): New population.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], numpy.ndarray, float]:
                1. New population.
                2. New populations fitness/function values.
                3. New global best individual.
                4. New global best fitness.

        """
        sparks_fitness = np.apply_along_axis(task.eval, 1, sparks)
        ib = np.argmin(sparks_fitness)
        best_x = sparks[ib].copy()
        best_fitness = sparks_fitness[ib]
        if sparks_fitness[ib] < population_fitness[0]:
            population[0], population_fitness[0] = best_x, best_fitness
        for i in range(1, self.population_size):
            r = self.integers(len(sparks))
            if sparks_fitness[r] < population_fitness[i]:
                population[i], population_fitness[i] = sparks[r], sparks_fitness[r]
        return population, population_fitness, best_x, best_fitness


class DynamicFireworksAlgorithmGauss(EnhancedFireworksAlgorithm):
    r"""Implementation of dynamic fireworks algorithm.

    Algorithm:
        Dynamic Fireworks Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        http://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=6900485&isnumber=6900223

    Reference paper:
        S. Zheng, A. Janecek, J. Li and Y. Tan, "Dynamic search in fireworks algorithm," 2014 IEEE Congress on Evolutionary Computation (CEC), Beijing, 2014, pp. 3222-3229. doi: 10.1109/CEC.2014.6900485

    Attributes:
        Name (List[str]): List of strings representing algorithm names.
        amplitude_cf (Union[float, int]): Amplitude of the core firework.
        amplification_coeff (Union[float, int]): Amplification coefficient.
        reduction_coeff (Union[float, int]): Reduction coefficient.

    """

    Name = ['DynamicFireworksAlgorithmGauss', 'dynFWAG']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""S. Zheng, A. Janecek, J. Li and Y. Tan, "Dynamic search in fireworks algorithm," 2014 IEEE Congress on Evolutionary Computation (CEC), Beijing, 2014, pp. 3222-3229. doi: 10.1109/CEC.2014.6900485"""

    def __init__(self, amplification_coeff=1.2, reduction_coeff=0.9, *args, **kwargs):
        """Initialize dynFWAG.

        Args:
            amplification_coeff (Union[int, float]): Amplification coefficient.
            reduction_coeff (Union[int, float]): Reduction coefficient.

        See Also:
            * :func:`FireworksAlgorithm.__init__`

        """
        super().__init__(*args, **kwargs)
        self.amplification_coeff = amplification_coeff
        self.reduction_coeff = reduction_coeff

    def set_parameters(self, amplification_coeff=1.2, reduction_coeff=0.9, **kwargs):
        r"""Set core arguments of DynamicFireworksAlgorithmGauss.

        Args:
            amplification_coeff (Union[int, float]): Amplification coefficient.
            reduction_coeff (Union[int, float]): Reduction coefficient.

        See Also:
            * :func:`FireworksAlgorithm.set_parameters`

        """
        super().set_parameters(**kwargs)
        self.amplification_coeff = amplification_coeff
        self.reduction_coeff = reduction_coeff

    def update_cf(self, xnb, xcb, xcb_f, xb, xb_f, amplitude_cf, task):
        r"""Update the core firework.

        Args:
            xnb: Sparks generated by core fireworks.
            xcb: Current generations best spark.
            xcb_f: Current generations best fitness.
            xb: Global best individual.
            xb_f: Global best fitness.
            amplitude_cf: Amplitude of the core firework.
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, float, numpy.ndarray]:
                1. New core firework.
                2. New core firework's fitness.
                3. New core firework amplitude.

        """
        xnb_f = np.apply_along_axis(task.eval, 1, xnb)
        ib_f = np.argmin(xnb_f)
        if xnb_f[ib_f] <= xb_f:
            xb, xb_f = xnb[ib_f], xnb_f[ib_f]
        if xb_f >= xcb_f:
            xb, xb_f, amplitude_cf = xcb, xcb_f, amplitude_cf * self.amplification_coeff
        else:
            amplitude_cf = amplitude_cf * self.reduction_coeff
        return xb, xb_f, amplitude_cf

    def explosion_amplitudes(self, population_fitness, task=None):
        """Calculate explosion amplitude for other fireworks."""
        return FireworksAlgorithm.explosion_amplitudes(self, population_fitness)

    def init_population(self, task):
        r"""Initialize population.

        Args:
            task (Task): Optimization task.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. Initialized population.
                2. Initialized population function/fitness values.
                3. Additional arguments:
                    * amplitude_cf (numpy.ndarray): Initial amplitude of the core firework.

        """
        fireworks, fitness, _ = super().init_population(task)
        amplitude_cf = task.range
        return fireworks, fitness, {'amplitude_cf': amplitude_cf}

    def selection(self, population, population_fitness, sparks, task):
        """Select fireworks for the next generation."""
        sparks_fitness = np.apply_along_axis(task.eval, 1, sparks)
        ib = np.argmin(sparks_fitness)
        for i, f in enumerate(population_fitness):
            r = self.integers(len(sparks))
            if sparks_fitness[r] < f:
                population[i], population_fitness[i] = sparks[r], sparks_fitness[r]
        population[0], population_fitness[0] = sparks[ib], sparks_fitness[ib]
        return population, population_fitness

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of DynamicFireworksAlgorithmGauss algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations function/fitness values.
            best_x (numpy.ndarray): Global best individual.
            best_fitness (float): Global best fitness/function value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population.
                2. New populations fitness/function values.
                3. New global best solution.
                4. New global best solutions fitness/objective value.
                5. Additional arguments:
                    * amplitude_cf (numpy.ndarray): Amplitude of the core firework.

        """
        amplitude_cf = params.pop('amplitude_cf')

        sparks_num = self.sparks_num(population_fitness)
        amplitudes = self.explosion_amplitudes(population_fitness)

        cf_sparks_num = self.num_sparks * (np.amax(population_fitness) - best_fitness + self.epsilon) /\
            (np.sum(np.amax(population_fitness - best_fitness)) + self.epsilon)

        if cf_sparks_num < self.a * self.num_sparks:
            cf_sparks_num = round(self.a * self.num_sparks)
        elif cf_sparks_num > self.b * self.num_sparks and self.a < self.b < 1:
            cf_sparks_num = round(self.b * self.num_sparks)
        else:
            cf_sparks_num = round(cf_sparks_num)

        all_sparks = population.copy()
        for i in range(self.population_size):
            si = sparks_num[i]
            ai = amplitudes[i]

            sparks_i = np.empty((si, task.dimension))
            for s in range(si):
                sparks_i[s] = population[i]
                sparks_i[s] = self.explosion_spark(sparks_i[s], ai, task)
            all_sparks = np.concatenate((all_sparks, sparks_i), axis=0)

        gaussian_idx = self.rng.choice(len(all_sparks), self.num_gaussian, replace=False)
        gaussian_sparks = np.array(all_sparks[gaussian_idx])

        for i in range(self.num_gaussian):
            gaussian_sparks[i] = self.gaussian_spark(gaussian_sparks[i], task, best_x=best_x)

        all_sparks = np.concatenate((all_sparks, gaussian_sparks), axis=0)

        cf_sparks = np.empty((cf_sparks_num, task.dimension))
        for s in range(cf_sparks_num):
            cf_sparks[s] = best_x
            cf_sparks[s] = self.explosion_spark(cf_sparks[s], amplitude_cf, task)

        population, population_fitness = self.selection(population, population_fitness, all_sparks, task)
        best_x, best_fitness, amplitude_cf = self.update_cf(cf_sparks, population[0], population_fitness[0], best_x, best_fitness, amplitude_cf, task)
        return population, population_fitness, best_x, best_fitness, {'amplitude_cf': amplitude_cf}


class DynamicFireworksAlgorithm(DynamicFireworksAlgorithmGauss):
    r"""Implementation of dynamic fireworks algorithm.

    Algorithm:
        Dynamic Fireworks Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        http://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=6900485&isnumber=6900223

    Reference paper:
        S. Zheng, A. Janecek, J. Li and Y. Tan, "Dynamic search in fireworks algorithm," 2014 IEEE Congress on Evolutionary Computation (CEC), Beijing, 2014, pp. 3222-3229. doi: 10.1109/CEC.2014.6900485

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.basic.DynamicFireworksAlgorithmGauss`

    """

    Name = ['DynamicFireworksAlgorithm', 'dynFWA']

    @staticmethod
    def info():
        r"""Get default information of algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""S. Zheng, A. Janecek, J. Li and Y. Tan, "Dynamic search in fireworks algorithm," 2014 IEEE Congress on Evolutionary Computation (CEC), Beijing, 2014, pp. 3222-3229. doi: 10.1109/CEC.2014.6900485"""

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Co50re function of Dynamic Fireworks Algorithm.

        Args:
            task (Task): Optimization task
            population (numpy.ndarray): Current population
            population_fitness (numpy.ndarray[float]): Current population fitness/function values
            best_x (numpy.ndarray): Current best solution
            best_fitness (float): Current best solution's fitness/function value
            **params:

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. New population.
                2. New population function/fitness values.
                3. New global best solution.
                4. New global best fitness.
                5. Additional arguments.

        """
        amplitude_cf = params.pop('amplitude_cf')

        sparks_num = self.sparks_num(population_fitness)
        amplitudes = self.explosion_amplitudes(population_fitness)

        cf_sparks_num = self.num_sparks * (np.amax(population_fitness) - best_fitness + self.epsilon) / \
            (np.sum(np.amax(population_fitness - best_fitness)) + self.epsilon)

        if cf_sparks_num < self.a * self.num_sparks:
            cf_sparks_num = round(self.a * self.num_sparks)
        elif cf_sparks_num > self.b * self.num_sparks and self.a < self.b < 1:
            cf_sparks_num = round(self.b * self.num_sparks)
        else:
            cf_sparks_num = round(cf_sparks_num)

        all_sparks = population.copy()
        for i in range(self.population_size):
            si = sparks_num[i]
            ai = amplitudes[i]

            sparks_i = np.empty((si, task.dimension))
            for s in range(si):
                sparks_i[s] = population[i]
                sparks_i[s] = self.explosion_spark(sparks_i[s], ai, task)
            all_sparks = np.concatenate((all_sparks, sparks_i), axis=0)

        cf_sparks = np.empty((cf_sparks_num, task.dimension))
        for s in range(cf_sparks_num):
            cf_sparks[s] = best_x
            cf_sparks[s] = self.explosion_spark(cf_sparks[s], amplitude_cf, task)

        population, population_fitness = self.selection(population, population_fitness, all_sparks, task)
        best_x, best_fitness, amplitude_cf = self.update_cf(cf_sparks, population[0], population_fitness[0], best_x,
                                                            best_fitness, amplitude_cf, task)
        return population, population_fitness, best_x, best_fitness, {'amplitude_cf': amplitude_cf}



# encoding=utf8

"""Implementations of Katsuura functions."""

from math import fabs
from niapy.problems.problem import Problem

__all__ = ['Katsuura']


class Katsuura(Problem):
    r"""Implementations of Katsuura functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
        **Katsuura Function**

        :math:`f(\textbf{x}) = \frac{10}{D^2} \prod_{i=1}^D \left( 1 + i \sum_{j=1}^{32} \frac{\lvert 2^j x_i - round\left(2^j x_i \right) \rvert}{2^j} \right)^\frac{10}{D^{1.2}} - \frac{10}{D^2}`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = \frac{10}{D^2} \prod_{i=1}^D \left( 1 + i \sum_{j=1}^{32} \frac{\lvert 2^j x_i - round\left(2^j x_i \right) \rvert}{2^j} \right)^\frac{10}{D^{1.2}} - \frac{10}{D^2}$
        Equation:
            \begin{equation} f(\textbf{x}) = \frac{10}{D^2} \prod_{i=1}^D \left( 1 + i \sum_{j=1}^{32} \frac{\lvert 2^j x_i - round\left(2^j x_i \right) \rvert}{2^j} \right)^\frac{10}{D^{1.2}} - \frac{10}{D^2} \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, dimension=5, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Katsuura problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = \frac{10}{D^2} \prod_{i=1}^D \left( 1 + i \sum_{j=1}^{32} \frac{| 2^j x_i - round\left(2^j x_i \right) |}{2^j} \right)^\frac{10}{D^{1.2}} - \frac{10}{D^2}$'''

    def _evaluate(self, x):
        val = 1.0
        for i in range(self.dimension):
            val_t = 1.0
            for j in range(1, 33):
                val_t += fabs(2 ** j * x[i] - round(2 ** j * x[i])) / 2 ** j
            val *= (1 + (i + 1) * val_t) ** (10 / self.dimension ** 1.2) - (10 / self.dimension ** 2)
        return 10 / self.dimension ** 2 * val



# encoding=utf8

"""The implementation of tasks."""

import logging
from enum import Enum

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.ticker as ticker
from niapy.problems import Problem
from niapy.util.repair import limit
from niapy.util.factory import get_problem

logging.basicConfig()
logger = logging.getLogger("niapy.task.Task")
logger.setLevel("INFO")


class OptimizationType(Enum):
    r"""Enum representing type of optimization.

    Attributes:
        MINIMIZATION (int): Represents minimization problems and is default optimization type of all algorithms.
        MAXIMIZATION (int): Represents maximization problems.

    """

    MINIMIZATION = 1.0
    MAXIMIZATION = -1.0


class Task:
    r"""Class representing an optimization task.

    Date:
        2019

    Author:
        Klemen Berkovič and others

    Attributes:
        problem (Problem): Optimization problem.
        dimension (int): Dimension of the problem.
        lower (numpy.ndarray): Lower bounds of the problem.
        upper (numpy.ndarray): Upper bounds of the problem.
        range (numpy.ndarray): Search range between upper and lower limits.
        optimization_type (OptimizationType): Optimization type to use.
        iters (int): Number of algorithm iterations/generations.
        evals (int): Number of function evaluations.
        max_iters (int): Maximum number of algorithm iterations/generations.
        max_evals (int): Maximum number of function evaluations.
        cutoff_value (float): Reference function/fitness values to reach in optimization.
        x_f (float): Best found individual function/fitness value.

    """

    def __init__(self, problem=None, dimension=None, lower=None, upper=None,
                 optimization_type=OptimizationType.MINIMIZATION, repair_function=limit, max_evals=np.inf,
                 max_iters=np.inf, cutoff_value=None, enable_logging=False):
        r"""Initialize task class for optimization.

        Args:
            problem (Union[str, Problem]): Optimization problem.
            dimension (Optional[int]): Dimension of the problem. Will be ignored if problem is instance of the `Problem` class.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem. Will be ignored if problem is instance of the `Problem` class.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem. Will be ignored if problem is instance of the `Problem` class.
            optimization_type (Optional[OptimizationType]): Set the type of optimization. Default is minimization.
            repair_function (Optional[Callable[[numpy.ndarray, numpy.ndarray, numpy.ndarray, Dict[str, Any]], numpy.ndarray]]): Function for repairing individuals components to desired limits.
            max_evals (Optional[int]): Number of function evaluations.
            max_iters (Optional[int]): Number of generations or iterations.
            cutoff_value (Optional[float]): Reference value of function/fitness function.
            enable_logging (Optional[bool]): Enable/disable logging of improvements.

        """
        if isinstance(problem, str):
            params = dict(dimension=dimension, lower=lower, upper=upper)
            params = {key: val for key, val in params.items() if val is not None}
            self.problem = get_problem(problem, **params)
        elif isinstance(problem, Problem):
            self.problem = problem
            if dimension is not None or lower is not None or upper is not None:
                logger.warning('An instance of the Problem class was passed in, `dimension`, `lower` and `upper` parameters will be ignored.')
        else:
            raise TypeError('Unsupported type for problem: {}'.format(type(problem)))

        self.optimization_type = optimization_type
        self.dimension = self.problem.dimension
        self.lower = self.problem.lower
        self.upper = self.problem.upper
        self.range = self.upper - self.lower
        self.repair_function = repair_function

        self.iters = 0
        self.evals = 0

        self.cutoff_value = -np.inf * optimization_type.value if cutoff_value is None else cutoff_value
        self.enable_logging = enable_logging
        self.x_f = np.inf * optimization_type.value
        self.max_evals = max_evals
        self.max_iters = max_iters
        self.n_evals = []
        self.fitness_evals = []  # fitness improvements at self.n_evals evaluations
        self.fitness_iters = []  # best fitness at each iteration

    def repair(self, x, rng=None):
        r"""Repair solution and put the solution in the random position inside of the bounds of problem.

        Args:
            x (numpy.ndarray): Solution to check and repair if needed.
            rng (Optional[numpy.random.Generator]): Random number generator.

        Returns:
            numpy.ndarray: Fixed solution.

        See Also:
            * :func:`niapy.util.repair.limit`
            * :func:`niapy.util.repair.limit_inverse`
            * :func:`niapy.util.repair.wang`
            * :func:`niapy.util.repair.rand`
            * :func:`niapy.util.repair.reflect`

        """
        return self.repair_function(x, self.lower, self.upper, rng=rng)

    def next_iter(self):
        r"""Increments the number of algorithm iterations."""
        self.fitness_iters.append(self.x_f)
        self.iters += 1

    def eval(self, x):
        r"""Evaluate the solution A.

        Args:
            x (numpy.ndarray): Solution to evaluate.

        Returns:
            float: Fitness/function values of solution.

        """
        if self.stopping_condition():
            return np.inf

        self.evals += 1
        x_f = self.problem.evaluate(x) * self.optimization_type.value

        if x_f < self.x_f * self.optimization_type.value:
            self.x_f = x_f * self.optimization_type.value
            self.n_evals.append(self.evals)
            self.fitness_evals.append(x_f)
            if self.enable_logging:
                logger.info('evals:%d => %s' % (self.evals, self.x_f))
        return x_f

    def is_feasible(self, x):
        r"""Check if the solution is feasible.

        Args:
            x (Union[numpy.ndarray, Individual]): Solution to check for feasibility.

        Returns:
            bool: `True` if solution is in feasible space else `False`.

        """
        return np.all((x >= self.lower) & (x <= self.upper))

    def stopping_condition(self):
        r"""Check if optimization task should stop.

        Returns:
            bool: `True` if number of function evaluations or number of algorithm iterations/generations or reference values is reach else `False`.

        """
        return (self.evals >= self.max_evals) or (self.iters >= self.max_iters) or (self.cutoff_value * self.optimization_type.value >= self.x_f * self.optimization_type.value)

    def stopping_condition_iter(self):
        r"""Check if stopping condition reached and increase number of iterations.

        Returns:
            bool: `True` if number of function evaluations or number of algorithm iterations/generations or reference values is reach else `False`.

        """
        r = self.stopping_condition()
        self.next_iter()
        return r

    def convergence_data(self, x_axis='iters'):
        r"""Get values of x and y-axis for plotting covariance graph.

        Args:
            x_axis (Literal['iters', 'evals']): Quantity to be displayed on the x-axis. Either 'iters' or 'evals'.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                1. array  of function evaluations.
                2. array of fitness values.

        """
        if x_axis == 'iters':
            return np.arange(self.iters), np.array(self.fitness_iters)
        else:  # x_axis == 'evals'
            r1, r2 = [], []
            for i, v in enumerate(self.n_evals):
                r1.append(v)
                r2.append(self.fitness_evals[i])
                if i >= len(self.n_evals) - 1:
                    break
                diff = self.n_evals[i + 1] - v
                if diff <= 1:
                    continue
                for j in range(diff - 1):
                    r1.append(v + j + 1)
                    r2.append(self.fitness_evals[i])
            return np.array(r1), np.array(r2)

    def plot_convergence(self, x_axis='iters', title='Convergence Graph'):
        """Plot a simple convergence graph.

        Args:
            x_axis (Literal['iters', 'evals']): Quantity to be displayed on the x-axis. Either 'iters' or 'evals'.
            title (str): Title of the graph.

        """
        x, fitness = self.convergence_data(x_axis)
        _, ax = plt.subplots()
        ax.plot(x, fitness)
        ax.xaxis.set_major_locator(ticker.MaxNLocator(integer=True))
        if x_axis == 'iters':
            plt.xlabel('Iterations')
        else:
            plt.xlabel('Fitness Evaluations')
        plt.ylabel('Fitness')
        plt.title(title)
        plt.show()



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm
from niapy.util.distances import euclidean

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['BacterialForagingOptimization']


class BacterialForagingOptimization(Algorithm):
    r"""Implementation of the Bacterial foraging optimization algorithm.

    Algorithm:
        Bacterial Foraging Optimization

    Date:
        2021

    Author:
        Žiga Stupan

    License:
        MIT

    Reference paper:
        K. M. Passino, "Biomimicry of bacterial foraging for distributed optimization and control," in IEEE Control Systems Magazine, vol. 22, no. 3, pp. 52-67, June 2002, doi: 10.1109/MCS.2002.1004010.

    Attributes:
        Name (List[str]): list of strings representing algorithm names.
        population_size (Optional[int]): Number of individuals in population :math:`\in [1, \infty]`.
        n_chemotactic (Optional[int]): Number of chemotactic steps.
        n_swim (Optional[int]): Number of swim steps.
        n_reproduction (Optional[int]): Number of reproduction steps.
        n_elimination (Optional[int]): Number of elimination and dispersal steps.
        prob_elimination (Optional[float]): Probability of a bacterium being eliminated and a new one being created at a random location in the search space.
        step_size (Optional[float]): Size of a chemotactic step.
        d_attract (Optional[float]): Depth of the attractant released by the cell (a quantification of how much attractant is released).
        w_attract (Optional[float]): Width of the attractant signal (a quantification of the diffusion rate of the chemical).
        h_repel (Optional[float]): Height of the repellent effect (magnitude of its effect).
        w_repel (Optional[float]): Width of the repellent.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['BacterialForagingOptimization', 'BFO', 'BFOA']

    @staticmethod
    def info():
        r"""Get algorithm information.

        Returns:
            str: Bit item.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""K. M. Passino, "Biomimicry of bacterial foraging for distributed optimization and control," in IEEE Control Systems Magazine, vol. 22, no. 3, pp. 52-67, June 2002, doi: 10.1109/MCS.2002.1004010."""

    def __init__(self, population_size=50, n_chemotactic=100, n_swim=4, n_reproduction=4, n_elimination=2,
                 prob_elimination=0.25, step_size=0.1, swarming=True, d_attract=0.1, w_attract=0.2, h_repel=0.1,
                 w_repel=10.0, *args, **kwargs):
        r"""Initialize algorithm.

        Args:
            population_size (Optional[int]): Number of individuals in population :math:`\in [1, \infty]`.
            n_chemotactic (Optional[int]): Number of chemotactic steps.
            n_swim (Optional[int]): Number of swim steps.
            n_reproduction (Optional[int]): Number of reproduction steps.
            n_elimination (Optional[int]): Number of elimination and dispersal steps.
            prob_elimination (Optional[float]): Probability of a bacterium being eliminated and a new one being created at a random location in the search space.
            step_size (Optional[float]): Size of a chemotactic step.
            swarming (Optional[bool]): If `True` use swarming.
            d_attract (Optional[float]): Depth of the attractant released by the cell (a quantification of how much attractant is released).
            w_attract (Optional[float]): Width of the attractant signal (a quantification of the diffusion rate of the chemical).
            h_repel (Optional[float]): Height of the repellent effect (magnitude of its effect).
            w_repel (Optional[float]): Width of the repellent.

        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.n_chemotactic = n_chemotactic
        self.n_swim = n_swim
        self.n_reproduction = n_reproduction
        self.n_elimination = n_elimination
        self.prob_elimination = prob_elimination
        self.step_size = step_size
        self.swarming = swarming
        self.d_attract = d_attract
        self.w_attract = w_attract
        self.h_repel = h_repel
        self.w_repel = w_repel

        self.i = 0  # elimination and dispersal step counter
        self.j = 0  # reproduction step counter
        self.k = 0  # chemotaxis step counter

    def set_parameters(self, population_size=50, n_chemotactic=100, n_swim=4, n_reproduction=4, n_elimination=2,
                       prob_elimination=0.25, step_size=0.1, swarming=True, d_attract=0.1, w_attract=0.2, h_repel=0.1,
                       w_repel=10.0, **kwargs):
        r"""Set the parameters/arguments of the algorithm.

        Args:
            population_size (Optional[int]): Number of individuals in population :math:`\in [1, \infty]`.
            n_chemotactic (Optional[int]): Number of chemotactic steps.
            n_swim (Optional[int]): Number of swim steps.
            n_reproduction (Optional[int]): Number of reproduction steps.
            n_elimination (Optional[int]): Number of elimination and dispersal steps.
            prob_elimination (Optional[float]): Probability of a bacterium being eliminated and a new one being created at a random location in the search space.
            step_size (Optional[float]): Size of a chemotactic step.
            swarming (Optional[bool]): If `True` use swarming.
            d_attract (Optional[float]): Depth of the attractant released by the cell (a quantification of how much attractant is released).
            w_attract (Optional[float]): Width of the attractant signal (a quantification of the diffusion rate of the chemical).
            h_repel (Optional[float]): Height of the repellent effect (magnitude of its effect).
            w_repel (Optional[float]): Width of the repellent.

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.n_chemotactic = n_chemotactic
        self.n_swim = n_swim
        self.n_reproduction = n_reproduction
        self.n_elimination = n_elimination
        self.prob_elimination = prob_elimination
        self.step_size = step_size
        self.d_attract = d_attract
        self.w_attract = w_attract
        self.h_repel = h_repel
        self.w_repel = w_repel

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        """
        params = super().get_parameters()
        params.update({
            'n_chemotactic': self.n_chemotactic,
            'n_swim': self.n_swim,
            'n_reproduction': self.n_reproduction,
            'n_elimination': self.n_elimination,
            'prob_elimination': self.prob_elimination,
            'step_size': self.step_size,
            'd_attract': self.d_attract,
            'w_attract': self.w_attract,
            'h_repel': self.h_repel,
            'w_repel': self.w_repel
        })
        return params

    def init_population(self, task):
        r"""Initialize the starting population.

        Args:
            task (Task): Optimization task

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, Dict[str, Any]]:
                1. New population.
                2. New population fitness/function values.
                3. Additional arguments:
                    * cost (numpy.ndarray): Costs of cells i. e. Fitness + cell interaction
                    * health (numpy.ndarray): Cell health i. e. The accumulation of costs over all chemotactic steps.

        See Also:
            * :func:`niapy.algorithms.Algorithm.init_population`

        """
        pop, fpop, d = super().init_population(task)
        d.update({
            'cost': np.zeros(self.population_size, dtype=np.float64),
            'health': np.zeros(self.population_size, dtype=np.float64)
        })
        return pop, fpop, d

    def interaction(self, cell, population):
        r"""Compute cell to cell interaction J_cc.

        Args:
            cell (numpy.ndarray): Cell to compute interaction for.
            population (numpy.ndarray): Population

        Returns:
            float: Cell to cell interaction J_cc

        """
        if not self.swarming:
            return 0.0
        distances = euclidean(cell, population)
        attract = np.sum(-self.d_attract * np.exp(-self.w_attract * distances))
        repel = np.sum(self.h_repel * np.exp(-self.w_repel * distances))
        return attract + repel

    def random_direction(self, dimension):
        r"""Generate a random direction vector.

        Args:
            dimension (int): Problem dimension

        Returns:
            numpy.ndarray: Normalised random direction vector

        """
        delta = self.uniform(-1.0, 1.0, dimension)
        return delta / np.linalg.norm(delta)

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Bacterial Foraging Optimization algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations fitness/function values.
            best_x (numpy.ndarray): Global best individual.
            best_fitness (float): Global best individuals function/fitness value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population.
                2. New populations function/fitness values.
                3. New global best solution,
                4. New global best solutions fitness/objective value.
                5. Additional arguments:
                    * cost (numpy.ndarray): Costs of cells i. e. Fitness + cell interaction
                    * health (numpy.ndarray): Cell health i. e. The accumulation of costs over all chemotactic steps.

        """
        cost = params.pop('cost')
        health = params.pop('health')

        # Chemotaxis
        for i in range(len(population)):
            cost[i] = population_fitness[i] + self.interaction(population[i], population)
            j_last = cost[i]
            step_direction = self.random_direction(task.dimension)

            m = 0
            while True:
                population[i] = task.repair(population[i] + self.step_size * step_direction)
                population_fitness[i] = task.eval(population[i])

                if population_fitness[i] < best_fitness:
                    best_x = population[i].copy()
                    best_fitness = population_fitness[i]

                cost[i] = population_fitness[i] + self.interaction(population[i], population)
                health[i] += cost[i]

                if m >= self.n_swim or cost[i] >= j_last:
                    break
                m += 1

        self.k += 1

        if self.k >= self.n_chemotactic:
            self.k = 0
            self.j += 1

            # Reproduction
            sorted_indices = np.argsort(health)
            population = population[sorted_indices]
            population_fitness = population_fitness[sorted_indices]
            cost = cost[sorted_indices]

            population = np.tile(population[:self.population_size // 2], (2, 1))
            population_fitness = np.tile(population_fitness[:self.population_size // 2], 2)
            cost = np.tile(cost[:self.population_size // 2], 2)

            health = np.zeros(len(population), dtype=np.float64)

        if self.j >= self.n_reproduction:
            self.j = 0
            self.i += 1

            # Elimination and dispersal
            for i in range(len(population)):
                if self.random() < self.prob_elimination:
                    population[i] = task.lower + self.random(task.dimension) * task.range
                    population_fitness[i] = task.eval(population[i])
                    if population_fitness[i] < best_fitness:
                        best_x = population[i].copy()
                        best_fitness = population_fitness[i]

        return population, population_fitness, best_x, best_fitness, {'cost': cost, 'health': health}



# encoding=utf8

"""Implementation of Rastrigin function."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Rastrigin']


class Rastrigin(Problem):
    r"""Implementation of Rastrigin problem.

    Date: 2018

    Authors: Lucija Brezočnik and Iztok Fister Jr.

    License: MIT

    Function: **Rastrigin function**

        :math:`f(\mathbf{x}) = 10D + \sum_{i=1}^D \left(x_i^2 -10\cos(2\pi x_i)\right)`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-5.12, 5.12]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x}) = 10D + \sum_{i=1}^D \left(x_i^2 -10\cos(2\pi x_i)\right)$

        Equation:
            \begin{equation} f(\mathbf{x}) =
            10D + \sum_{i=1}^D \left(x_i^2 -10\cos(2\pi x_i)\right)
            \end{equation}

        Domain:
            $-5.12 \leq x_i \leq 5.12$

    Reference:
        https://www.sfu.ca/~ssurjano/rastr.html

    """

    def __init__(self, dimension=4, lower=-5.12, upper=5.12, *args, **kwargs):
        r"""Initialize Rastrigin problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\mathbf{x}) = 10D + \sum_{i=1}^D \left(x_i^2 -10\cos(2\pi x_i)\right)$'''

    def _evaluate(self, x):
        return 10.0 * self.dimension + np.sum(x * x - 10.0 * np.cos(2 * np.pi * x))



# encoding=utf8

"""Implementation of Pinter function."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Pinter']


class Pinter(Problem):
    r"""Implementation of Pintér function.

    Date: 2018

    Author: Lucija Brezočnik

    License: MIT

    Function: **Pintér function**

        :math:`f(\mathbf{x}) =
        \sum_{i=1}^D ix_i^2 + \sum_{i=1}^D 20i \sin^2 A + \sum_{i=1}^D i
        \log_{10} (1 + iB^2);`
        :math:`A = (x_{i-1}\sin(x_i)+\sin(x_{i+1}))\quad \text{and} \quad`
        :math:`B = (x_{i-1}^2 - 2x_i + 3x_{i+1} - \cos(x_i) + 1)`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-10, 10]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x}) =
            \sum_{i=1}^D ix_i^2 + \sum_{i=1}^D 20i \sin^2 A + \sum_{i=1}^D i
            \log_{10} (1 + iB^2);
            A = (x_{i-1}\sin(x_i)+\sin(x_{i+1}))\quad \text{and} \quad
            B = (x_{i-1}^2 - 2x_i + 3x_{i+1} - \cos(x_i) + 1)$

        Equation:
            \begin{equation} f(\mathbf{x}) = \sum_{i=1}^D ix_i^2 +
            \sum_{i=1}^D 20i \sin^2 A + \sum_{i=1}^D i \log_{10} (1 + iB^2);
            A = (x_{i-1}\sin(x_i)+\sin(x_{i+1}))\quad \text{and} \quad
            B = (x_{i-1}^2 - 2x_i + 3x_{i+1} - \cos(x_i) + 1) \end{equation}

        Domain:
            $-10 \leq x_i \leq 10$

    Reference paper:
        Jamil, M., and Yang, X. S. (2013).
        A literature survey of benchmark functions for global optimisation problems.
        International Journal of Mathematical Modelling and Numerical Optimisation,
        4(2), 150-194.

    """

    def __init__(self, dimension=4, lower=-10.0, upper=10.0, *args, **kwargs):
        r"""Initialize Pinter problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r''' $f(\mathbf{x}) =
                \sum_{i=1}^D ix_i^2 + \sum_{i=1}^D 20i \sin^2 A + \sum_{i=1}^D i
                \log_{10} (1 + iB^2);
                A = (x_{i-1}\sin(x_i)+\sin(x_{i+1}))\quad \text{and} \quad
                B = (x_{i-1}^2 - 2x_i + 3x_{i+1} - \cos(x_i) + 1)$'''

    def _evaluate(self, x):
        sub = np.roll(x, 1)
        add = np.roll(x, - 1)
        indices = np.arange(1, self.dimension + 1)

        a = (sub * np.sin(x) + np.sin(add))
        b = ((sub * sub) - 2.0 * x + 3.0 * add - np.cos(x) + 1.0)

        val1 = np.sum(indices * x * x)
        val2 = np.sum(20.0 * indices * np.power(np.sin(a), 2.0))
        val3 = np.sum(indices * np.log10(1.0 + indices * np.power(b, 2.0)))

        return val1 + val2 + val3



# encoding=utf8

"""Implementation of Runner utility class."""

import datetime
import logging
import os

import pandas as pd

from niapy.algorithms.algorithm import Algorithm
from niapy.task import Task
from niapy.util.factory import get_algorithm

logging.basicConfig()
logger = logging.getLogger('niapy.runner.Runner')
logger.setLevel('INFO')

__all__ = ["Runner"]


class Runner:
    r"""Runner utility feature.

    Feature which enables running multiple algorithms with multiple problems.
    It also support exporting results in various formats (e.g. Pandas DataFrame, JSON, Excel)

    Attributes:
        dimension (int): Dimension of problem
        max_evals (int): Number of function evaluations
        runs (int): Number of repetitions
        algorithms (Union[List[str], List[Algorithm]]): List of algorithms to run
        problems (List[Union[str, Problem]]): List of problems to run

    """

    def __init__(self, dimension=10, max_evals=1000000, runs=1, algorithms='ArtificialBeeColonyAlgorithm',
                 problems='Ackley'):
        r"""Initialize Runner.

        Args:
            dimension (int): Dimension of problem
            max_evals (int): Number of function evaluations
            runs (int): Number of repetitions
            algorithms (List[Algorithm]): List of algorithms to run
            problems (List[Union[str, Problem]]): List of problems to run

        """
        self.dimension = dimension
        self.max_evals = max_evals
        self.runs = runs
        self.algorithms = algorithms
        self.problems = problems
        self.results = {}

    def task_factory(self, name):
        r"""Create optimization task.

        Args:
            name (str): Problem name.

        Returns:
            Task: Optimization task to use.

        """
        return Task(max_evals=self.max_evals, dimension=self.dimension, problem=name)

    @classmethod
    def __create_export_dir(cls):
        if not os.path.exists("export"):
            os.makedirs("export")

    @classmethod
    def __generate_export_name(cls, extension):
        Runner.__create_export_dir()
        return "export/" + str(datetime.datetime.now()).replace(":", ".") + "." + extension

    def __export_to_dataframe_pickle(self):
        dataframe = pd.DataFrame.from_dict(self.results)
        dataframe.to_pickle(self.__generate_export_name("pkl"))
        logger.info("Export to Pandas DataFrame pickle (pkl) completed!")

    def __export_to_json(self):
        dataframe = pd.DataFrame.from_dict(self.results)
        dataframe.to_json(self.__generate_export_name("json"))
        logger.info("Export to JSON file completed!")

    def _export_to_xls(self):
        dataframe = pd.DataFrame.from_dict(self.results)
        dataframe.to_excel(self.__generate_export_name("xls"))
        logger.info("Export to XLS completed!")

    def __export_to_xlsx(self):
        dataframe = pd.DataFrame.from_dict(self.results)
        dataframe.to_excel(self.__generate_export_name("xslx"))
        logger.info("Export to XLSX file completed!")

    def run(self, export="dataframe", verbose=False):
        """Execute runner.

        Args:
            export (str): Takes export type (e.g. dataframe, json, xls, xlsx) (default: "dataframe")
            verbose (bool): Switch for verbose logging (default: {False})

        Returns:
            dict: Returns dictionary of results

        Raises:
            TypeError: Raises TypeError if export type is not supported

        """
        for alg in self.algorithms:
            if not isinstance(alg, "".__class__):
                alg_name = str(type(alg).__name__)
            else:
                alg_name = alg

            self.results[alg_name] = {}

            if verbose:
                logger.info("Running %s...", alg_name)

            for problem in self.problems:
                if not isinstance(problem, "".__class__):
                    problem_name = str(type(problem).__name__)
                else:
                    problem_name = problem

                if verbose:
                    logger.info("Running %s algorithm on %s problem...", alg_name, problem_name)

                self.results[alg_name][problem_name] = []
                for _ in range(self.runs):
                    if isinstance(alg, Algorithm):
                        algorithm = alg
                    else:
                        algorithm = get_algorithm(alg)
                    task = self.task_factory(problem)
                    self.results[alg_name][problem_name].append(algorithm.run(task))
            if verbose:
                logger.info("---------------------------------------------------")
        if export == "dataframe":
            self.__export_to_dataframe_pickle()
        elif export == "json":
            self.__export_to_json()
        elif export == "xsl":
            self._export_to_xls()
        elif export == "xlsx":
            self.__export_to_xlsx()
        else:
            raise TypeError("Passed export type %s is not supported!", export)
        return self.results



# encoding=utf8

"""Implementations of Cosine mixture functions."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['CosineMixture']


class CosineMixture(Problem):
    r"""Implementations of Cosine mixture function.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Cosine Mixture Function**

        :math:`f(\textbf{x}) = - 0.1 \sum_{i = 1}^D \cos (5 \pi x_i) - \sum_{i = 1}^D x_i^2`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-1, 1]`, for all :math:`i = 1, 2,..., D`.

        **Global maximum:**
        :math:`f(x^*) = -0.1 D`, at :math:`x^* = (0.0,...,0.0)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = - 0.1 \sum_{i = 1}^D \cos (5 \pi x_i) - \sum_{i = 1}^D x_i^2$

        Equation:
            \begin{equation} f(\textbf{x}) = - 0.1 \sum_{i = 1}^D \cos (5 \pi x_i) - \sum_{i = 1}^D x_i^2 \end{equation}

        Domain:
            $-1 \leq x_i \leq 1$

    Reference:
        http://infinity77.net/global_optimization/test_functions_nd_C.html#go_benchmark.CosineMixture

    """

    def __init__(self, dimension=4, lower=-1.0, upper=1.0, *args, **kwargs):
        r"""Initialize Cosine mixture problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = - 0.1 \sum_{i = 1}^D \cos (5 \pi x_i) - \sum_{i = 1}^D x_i^2$'''

    def _evaluate(self, x):
        return -0.1 * np.sum(np.cos(5 * np.pi * x)) - np.sum(x ** 2)



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm

logging.basicConfig()
logger = logging.getLogger("niapy.algorithms.basic")
logger.setLevel("INFO")

__all__ = ["HarmonySearch", "HarmonySearchV1"]


class HarmonySearch(Algorithm):
    r"""Implementation of Harmony Search algorithm.

    Algorithm:
        Harmony Search Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://journals.sagepub.com/doi/10.1177/003754970107600201

    Reference paper:
        Geem, Z. W., Kim, J. H., & Loganathan, G. V. (2001). A new heuristic optimization algorithm: harmony search. Simulation, 76(2), 60-68.

    Attributes:
        Name (List[str]): List of strings representing algorithm names
        r_accept (float): Probability of accepting new bandwidth into harmony.
        r_pa (float): Probability of accepting random bandwidth into harmony.
        b_range (float): Range of bandwidth.

    See Also:
        * :class:`niapy.algorithms.algorithm.Algorithm`

    """

    Name = ["HarmonySearch", "HS"]

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        """
        return r"""Geem, Z. W., Kim, J. H., & Loganathan, G. V. (2001). A new heuristic optimization algorithm: harmony search. Simulation, 76(2), 60-68."""

    def __init__(self, population_size=30, r_accept=0.7, r_pa=0.35, b_range=1.42, *args, **kwargs):
        """Initialize HarmonySearch.

        Args:
            population_size (Optional[int]): Number of harmony in the memory.
            r_accept (Optional[float]): Probability of accepting new bandwidth to harmony.
            r_pa (Optional[float]): Probability of accepting random bandwidth into harmony.
            b_range (Optional[float]): Bandwidth range.

        """
        super().__init__(population_size, *args, **kwargs)
        self.r_accept = r_accept
        self.r_pa = r_pa
        self.b_range = b_range

    def set_parameters(self, population_size=30, r_accept=0.7, r_pa=0.35, b_range=1.42, **kwargs):
        r"""Set the arguments of the algorithm.

        Args:
            population_size (Optional[int]): Number of harmony in the memory.
            r_accept (Optional[float]): Probability of accepting new bandwidth to harmony.
            r_pa (Optional[float]): Probability of accepting random bandwidth into harmony.
            b_range (Optional[float]): Bandwidth range.

        See Also:
            * :func:`niapy.algorithms.algorithm.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.r_accept = r_accept
        self.r_pa = r_pa
        self.b_range = b_range

    def get_parameters(self):
        """Get algorithm parameters."""
        d = super().get_parameters()
        d.update({
            'r_accept': self.r_accept,
            'r_pa': self.r_pa,
            'b_range': self.b_range
        })
        return d

    def bw(self, task):
        r"""Get bandwidth.

        Args:
            task (Task): Optimization task.

        Returns:
            float: Bandwidth.

        """
        return self.uniform(-1, 1) * self.b_range

    def adjustment(self, x, task):
        r"""Adjust value based on bandwidth.

        Args:
            x (Union[int, float]): Current position.
            task (Task): Optimization task.

        Returns:
            float: New position.

        """
        return x + self.bw(task)

    def improvise(self, harmonies, task):
        r"""Create new individual.

        Args:
            harmonies (numpy.ndarray): Current population.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray: New individual.

        """
        harmony = np.zeros(task.dimension)
        for i in range(task.dimension):
            r, j = self.random(), self.integers(self.population_size)
            harmony[i] = harmonies[j, i] if r > self.r_accept else self.adjustment(harmonies[j, i], task) if r > self.r_pa else self.uniform(
                task.lower[i], task.upper[i])
        return harmony

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of HarmonySearch algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population.
            population_fitness (numpy.ndarray): Current populations function/fitness values.
            best_x (numpy.ndarray): Global best individual.
            best_fitness (float): Global best fitness/function value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New harmony/population.
                2. New populations function/fitness values.
                3. New global best solution
                4. New global best solution fitness/objective value
                5. Additional arguments.

        """
        harmony = self.improvise(population, task)
        harmony_fitness = task.eval(task.repair(harmony, self.rng))
        iw = np.argmax(population_fitness)
        if harmony_fitness <= population_fitness[iw]:
            population[iw], population_fitness[iw] = harmony, harmony_fitness
        best_x, best_fitness = self.get_best(harmony, harmony_fitness, best_x, best_fitness)
        return population, population_fitness, best_x, best_fitness, {}


class HarmonySearchV1(HarmonySearch):
    r"""Implementation of harmony search algorithm.

    Algorithm:
        Harmony Search Algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://link.springer.com/chapter/10.1007/978-3-642-00185-7_1

    Reference paper:
        Yang, Xin-She. "Harmony search as a metaheuristic algorithm." Music-inspired harmony search algorithm. Springer, Berlin, Heidelberg, 2009. 1-14.

    Attributes:
        Name (List[str]): List of strings representing algorithm name.
        bw_min (float): Minimal bandwidth.
        bw_max (float): Maximal bandwidth.

    See Also:
        * :class:`niapy.algorithms.basic.hs.HarmonySearch`

    """

    Name = ["HarmonySearchV1", "HSv1"]

    @staticmethod
    def info():
        r"""Get basic information about algorithm.

        Returns:
            str: Basic information.

        """
        return r"""Yang, Xin-She. "Harmony search as a metaheuristic algorithm." Music-inspired harmony search algorithm. Springer, Berlin, Heidelberg, 2009. 1-14."""

    def __init__(self, bw_min=1, bw_max=2, *args, **kwargs):
        """Initialize HarmonySearchV1.

        Args:
            bw_min (Optional[float]): Minimal bandwidth.
            bw_max (Optional[float]): Maximal bandwidth.

        """
        super().__init__(*args, **kwargs)
        self.bw_min = bw_min
        self.bw_max = bw_max

    def set_parameters(self, bw_min=1, bw_max=2, **kwargs):
        r"""Set the parameters of the algorithm.

        Args:
            bw_min (Optional[float]): Minimal bandwidth
            bw_max (Optional[float]): Maximal bandwidth

        See Also:
            * :func:`niapy.algorithms.basic.hs.HarmonySearch.set_parameters`

        """
        super().set_parameters(**kwargs)
        self.bw_min, self.bw_max = bw_min, bw_max

    def get_parameters(self):
        """Get algorithm parameters."""
        d = super().get_parameters()
        d.update({
            'bw_min': self.bw_min,
            'bw_max': self.bw_max
        })
        return d

    def bw(self, task):
        r"""Get new bandwidth.

        Args:
            task (Task): Optimization task.

        Returns:
            float: New bandwidth.

        """
        return self.bw_min * np.exp(np.log(self.bw_min / self.bw_max) * (task.iters + 1) / task.max_iters)



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm
from niapy.util import levy_flight

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['HarrisHawksOptimization']


class HarrisHawksOptimization(Algorithm):
    r"""Implementation of Harris Hawks Optimization algorithm.

    Algorithm:
        Harris Hawks Optimization

    Date:
        2020

    Authors:
        Francisco Jose Solis-Munoz

    License:
        MIT

    Reference paper:
        Heidari et al. "Harris hawks optimization: Algorithm and applications". Future Generation Computer Systems. 2019. Vol. 97. 849-872.

    Attributes:
        Name (List[str]): List of strings representing algorithm name.
        levy (float): Levy factor.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['HarrisHawksOptimization', 'HHO']

    def __init__(self, population_size=40, levy=0.01, *args, **kwargs):
        """Initialize HarrisHawksOptimization.

        Args:
            population_size (Optional[int]): Population size.
            levy (Optional[float]): Levy factor.

        """
        super().__init__(population_size, *args, **kwargs)
        self.levy = levy

    @staticmethod
    def info():
        r"""Get algorithms information.

        Returns:
            str: Algorithm information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""Heidari et al. "Harris hawks optimization: Algorithm and applications". Future Generation Computer Systems. 2019. Vol. 97. 849-872."""

    def set_parameters(self, population_size=40, levy=0.01, **kwargs):
        r"""Set the parameters of the algorithm.

        Args:
            population_size (Optional[int]): Population size.
            levy (Optional[float]): Levy factor.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.levy = levy

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        """
        d = super().get_parameters()
        d.update({
            'levy': self.levy
        })
        return d

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Harris Hawks Optimization.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population
            population_fitness (numpy.ndarray[float]): Current population fitness/function values
            best_x (numpy.ndarray): Current best individual
            best_fitness (float): Current best individual function/fitness value
            params (Dict[str, Any]): Additional algorithm arguments

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population
                2. New population fitness/function values
                3. New global best solution
                4. New global best fitness/objective value

        """
        # Decreasing energy factor
        decreasing_energy_factor = 2 * (1 - (task.iters + 1) / task.max_iters)
        mean_sol = np.mean(population)
        # Update population
        for i in range(self.population_size):
            jumping_energy = self.rng.uniform(0, 2)
            decreasing_energy_random = self.rng.uniform(-1, 1)
            escaping_energy = decreasing_energy_factor * decreasing_energy_random
            escaping_energy_abs = np.abs(escaping_energy)
            random_number = self.rng.random()
            if escaping_energy >= 1 and random_number >= 0.5:
                # 0. Exploration: Random tall tree
                rhi = self.rng.integers(self.population_size)
                random_agent = population[rhi]
                population[i] = random_agent - self.rng.random() * np.abs(random_agent - 2 * self.rng.random() * population[i])
            elif escaping_energy_abs >= 1 and random_number < 0.5:
                # 1. Exploration: Family members mean
                population[i] = (best_x - mean_sol) - self.rng.random() * self.rng.uniform(task.lower, task.upper)
            elif escaping_energy_abs >= 0.5 and random_number >= 0.5:
                # 2. Exploitation: Soft besiege
                population[i] = \
                    (best_x - population[i]) - \
                    escaping_energy * \
                    np.abs(jumping_energy * best_x - population[i])
            elif escaping_energy_abs < 0.5 <= random_number:
                # 3. Exploitation: Hard besiege
                population[i] = \
                    best_x - \
                    escaping_energy * \
                    np.abs(best_x - population[i])
            elif escaping_energy_abs >= 0.5 > random_number:
                # 4. Exploitation: Soft besiege with progressive rapid dives
                cand1 = task.repair(best_x - escaping_energy * np.abs(jumping_energy * best_x - population[i]), rng=self.rng)
                random_vector = self.rng.random(task.dimension)
                cand2 = task.repair(cand1 + random_vector * levy_flight(alpha=self.levy, size=task.dimension, rng=self.rng),
                                    rng=self.rng)
                if task.eval(cand1) < population_fitness[i]:
                    population[i] = cand1
                elif task.eval(cand2) < population_fitness[i]:
                    population[i] = cand2
            elif escaping_energy_abs < 0.5 and random_number < 0.5:
                # 5. Exploitation: Hard besiege with progressive rapid dives
                cand1 = task.repair(best_x - escaping_energy * np.abs(jumping_energy * best_x - mean_sol), rng=self.rng)
                random_vector = self.rng.random(task.dimension)
                cand2 = task.repair(cand1 + random_vector * levy_flight(alpha=self.levy, size=task.dimension, rng=self.rng),
                                    rng=self.rng)
                if task.eval(cand1) < population_fitness[i]:
                    population[i] = cand1
                elif task.eval(cand2) < population_fitness[i]:
                    population[i] = cand2
            # Repair agent (from population) values
            population[i] = task.repair(population[i], rng=self.rng)
            # Eval population
            population_fitness[i] = task.eval(population[i])
        # Get best of population
        best_index = np.argmin(population_fitness)
        xb_cand = population[best_index].copy()
        fxb_cand = population_fitness[best_index].copy()
        if fxb_cand < best_fitness:
            best_fitness = fxb_cand
            best_x = xb_cand.copy()
        return population, population_fitness, best_x, best_fitness, {}



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Individual
from niapy.algorithms.basic.de import MultiStrategyDifferentialEvolution, DynNpDifferentialEvolution, DifferentialEvolution
from niapy.algorithms.other.mts import mts_ls1v1, mts_ls2, mts_ls3v1, MultipleTrajectorySearch

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.modified')
logger.setLevel('INFO')

__all__ = ['DifferentialEvolutionMTS', 'DifferentialEvolutionMTSv1', 'DynNpDifferentialEvolutionMTS',
           'DynNpDifferentialEvolutionMTSv1', 'MultiStrategyDifferentialEvolutionMTS',
           'MultiStrategyDifferentialEvolutionMTSv1', 'DynNpMultiStrategyDifferentialEvolutionMTS',
           'DynNpMultiStrategyDifferentialEvolutionMTSv1']


class MtsIndividual(Individual):
    r"""Individual for MTS local searches.

    Attributes:
        search_range (numpy.ndarray): Search range.
        grade (int): Grade of individual.
        enable (bool): If enabled.
        improved (bool): If improved.

    See Also:
        :class:`niapy.algorithms.algorithm.Individual`

    """

    def __init__(self, search_range=None, grade=0, enable=True, improved=False, task=None, **kwargs):
        r"""Initialize the individual.

        Args:
            search_range (numpy.ndarray): Search range.
            grade (Optional[int]): Grade of individual.
            enable (Optional[bool]): If enabled individual.
            improved (Optional[bool]): If individual improved.

        See Also:
            :func:`niapy.algorithms.algorithm.Individual.__init__`

        """
        super().__init__(task=task, **kwargs)
        self.grade, self.enable, self.improved = grade, enable, improved
        if search_range is None and task is not None:
            self.search_range = task.range / 4
        else:
            self.search_range = search_range


class DifferentialEvolutionMTS(DifferentialEvolution, MultipleTrajectorySearch):
    r"""Implementation of Differential Evolution with MTS local searches.

    Algorithm:
        Differential Evolution with MTS local searches

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm names.

    See Also:
        * :class:`niapy.algorithms.basic.de.DifferentialEvolution`
        * :class:`niapy.algorithms.other.mts.MultipleTrajectorySearch`

    """

    Name = ['DifferentialEvolutionMTS', 'DEMTS']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, population_size=40, *args, **kwargs):
        """Initialize DifferentialEvolutionMTS."""
        super().__init__(population_size, individual_type=kwargs.pop('individual_type', MtsIndividual), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set the algorithm parameters.

        See Also:
            :func:`niapy.algorithms.basic.de.DifferentialEvolution.set_parameters`

        """
        MultipleTrajectorySearch.set_parameters(self, **kwargs)
        DifferentialEvolution.set_parameters(self, individual_type=kwargs.pop('individual_type', MtsIndividual),
                                             **kwargs)

    def get_parameters(self):
        """Get algorithm parameters."""
        d = DifferentialEvolution.get_parameters(self)
        d.update(MultipleTrajectorySearch.get_parameters(self))
        return d

    def post_selection(self, population, task, xb, fxb, **kwargs):
        r"""Post selection operator.

        Args:
            population (numpy.ndarray): Current population.
            task (Task): Optimization task.
            xb (numpy.ndarray): Global best individual.
            fxb (float): Global best fitness.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, float]: New population.

        """
        for x in population:
            if not x.enable:
                continue
            x.enable, x.grades = False, 0
            x.x, x.f, xb, fxb, k = self.grading_run(x.x, x.f, xb, fxb, x.improved, x.search_range, task)
            x.x, x.f, xb, fxb, x.improved, x.search_range, x.grades = self.run_local_search(k, x.x, x.f, xb, fxb, x.improved, x.search_range, 0, task)
        for i in population[np.argsort([x.grade for x in population])[:self.num_enabled:]]:
            i.enable = True
        return population, xb, fxb


class DifferentialEvolutionMTSv1(DifferentialEvolutionMTS):
    r"""Implementation of Differential Evolution with MTSv1 local searches.

    Algorithm:
        Differential Evolution with MTSv1 local searches

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        :class:`niapy.algorithms.modified.DifferentialEvolutionMTS`

    """

    Name = ['DifferentialEvolutionMTSv1', 'DEMTSv1']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize DifferentialEvolutionMTSv1."""
        super().__init__(local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core parameters of DifferentialEvolutionMTSv1 algorithm.

        See Also:
            :func:`niapy.algorithms.modified.DifferentialEvolutionMTS.set_parameters`

        """
        super().set_parameters(local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), **kwargs)


class DynNpDifferentialEvolutionMTS(DifferentialEvolutionMTS, DynNpDifferentialEvolution):
    r"""Implementation of Differential Evolution with MTS local searches dynamic and population size.

    Algorithm:
        Differential Evolution with MTS local searches and dynamic population size

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name

    See Also:
        * :class:`niapy.algorithms.modified.DifferentialEvolutionMTS`
        * :class:`niapy.algorithms.basic.de.DynNpDifferentialEvolution`

    """

    Name = ['DynNpDifferentialEvolutionMTS', 'dynNpDEMTS']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize DynNpDifferentialEvolutionMTS."""
        super().__init__(*args, **kwargs)

    def set_parameters(self, p_max=10, rp=3, **kwargs):
        r"""Set core parameters or DynNpDifferentialEvolutionMTS algorithm.

        Args:
            p_max (Optional[int]):
            rp (Optional[float]):

        See Also:
            * :func:`niapy.algorithms.modified.hde.DifferentialEvolutionMTS.set_parameters`
            * :func`niapy.algorithms.basic.de.DynNpDifferentialEvolution.set_parameters`

        """
        DynNpDifferentialEvolution.set_parameters(self, p_max=p_max, rp=rp, **kwargs)
        DifferentialEvolutionMTS.set_parameters(self, **kwargs)

    def post_selection(self, population, task, xb, fxb, **kwargs):
        new_x, xb, fxb = DynNpDifferentialEvolution.post_selection(self, population, task, xb, fxb)
        new_x, xb, fxb = DifferentialEvolutionMTS.post_selection(self, new_x, task, xb, fxb)
        return new_x, xb, fxb


class DynNpDifferentialEvolutionMTSv1(DynNpDifferentialEvolutionMTS):
    r"""Implementation of Differential Evolution with MTSv1 local searches and dynamic population size.

    Algorithm:
        Differential Evolution with MTSv1 local searches and dynamic population size

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        :class:`niapy.algorithms.modified.hde.DifferentialEvolutionMTS`

    """

    Name = ['DynNpDifferentialEvolutionMTSv1', 'dynNpDEMTSv1']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize DynNpDifferentialEvolutionMTSv1."""
        super().__init__(local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core arguments of DynNpDifferentialEvolutionMTSv1 algorithm.

        See Also:
            :func:`niapy.algorithms.modified.hde.DifferentialEvolutionMTS.set_parameters`

        """
        DynNpDifferentialEvolutionMTS.set_parameters(self, local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), **kwargs)


class MultiStrategyDifferentialEvolutionMTS(DifferentialEvolutionMTS, MultiStrategyDifferentialEvolution):
    r"""Implementation of Differential Evolution with MTS local searches and multiple mutation strategies.

    Algorithm:
        Differential Evolution with MTS local searches and multiple mutation strategies

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.modified.hde.DifferentialEvolutionMTS`
        * :class:`niapy.algorithms.basic.de.MultiStrategyDifferentialEvolution`

    """

    Name = ['MultiStrategyDifferentialEvolutionMTS', 'MSDEMTS']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize MultiStrategyDifferentialEvolutionMTS."""
        super().__init__(individual_type=kwargs.pop('individual_type', MtsIndividual), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set algorithm parameters.

        See Also:
            * :func:`niapy.algorithms.modified.DifferentialEvolutionMTS.set_parameters`
            * :func:`niapy.algorithms.basic.MultiStrategyDifferentialEvolution.set_parameters`

        """
        DifferentialEvolutionMTS.set_parameters(self, **kwargs)
        MultiStrategyDifferentialEvolution.set_parameters(self, individual_type=kwargs.pop('individual_type', MtsIndividual), **kwargs)

    def evolve(self, pop, xb, task, **kwargs):
        r"""Evolve population.

        Args:
            pop (numpy.ndarray[Individual]): Current population of individuals.
            xb (numpy.ndarray): Global best individual.
            task (Task): Optimization task.

        Returns:
            numpy.ndarray[Individual]: Evolved population.

        """
        return MultiStrategyDifferentialEvolution.evolve(self, pop, xb, task, **kwargs)


class MultiStrategyDifferentialEvolutionMTSv1(MultiStrategyDifferentialEvolutionMTS):
    r"""Implementation of Differential Evolution with MTSv1 local searches and multiple mutation strategies.

    Algorithm:
        Differential Evolution with MTSv1 local searches and multiple mutation strategies

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of stings representing algorithm name.

    See Also:
        * :class:`niapy.algorithms.modified.MultiStrategyDifferentialEvolutionMTS`

    """

    Name = ['MultiStrategyDifferentialEvolutionMTSv1', 'MSDEMTSv1']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize MultiStrategyDifferentialEvolutionMTSv1."""
        super().__init__(local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core parameters of MultiStrategyDifferentialEvolutionMTSv1 algorithm.

        See Also:
            * :func:`niapy.algorithms.modified.MultiStrategyDifferentialEvolutionMTS.set_parameters`

        """
        MultiStrategyDifferentialEvolutionMTS.set_parameters(self, local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), **kwargs)


class DynNpMultiStrategyDifferentialEvolutionMTS(MultiStrategyDifferentialEvolutionMTS, DynNpDifferentialEvolutionMTS):
    r"""Implementation of Differential Evolution with MTS local searches, multiple mutation strategies and dynamic population size.

    Algorithm:
        Differential Evolution with MTS local searches, multiple mutation strategies and dynamic population size

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name

    See Also:
        * :class:`niapy.algorithms.modified.MultiStrategyDifferentialEvolutionMTS`
        * :class:`niapy.algorithms.modified.DynNpDifferentialEvolutionMTS`

    """

    Name = ['DynNpMultiStrategyDifferentialEvolutionMTS', 'dynNpMSDEMTS']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize DynNpMultiStrategyDifferentialEvolutionMTS."""
        super().__init__(*args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core arguments of DynNpMultiStrategyDifferentialEvolutionMTS algorithm.

        See Also:
            * :func:`niapy.algorithms.modified.MultiStrategyDifferentialEvolutionMTS.set_parameters`
            * :func:`niapy.algorithms.modified.DynNpDifferentialEvolutionMTS.set_parameters`

        """
        DynNpDifferentialEvolutionMTS.set_parameters(self, **kwargs)
        MultiStrategyDifferentialEvolutionMTS.set_parameters(self, **kwargs)


class DynNpMultiStrategyDifferentialEvolutionMTSv1(DynNpMultiStrategyDifferentialEvolutionMTS):
    r"""Implementation of Differential Evolution with MTSv1 local searches, multiple mutation strategies and dynamic population size.

    Algorithm:
        Differential Evolution with MTSv1 local searches, multiple mutation strategies and dynamic population size

    Date:
        2018

    Author:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        Name (List[str]): List of strings representing algorithm name.

    See Also:
        * :class:`niapy.algorithm.modified.DynNpMultiStrategyDifferentialEvolutionMTS`

    """

    Name = ['DynNpMultiStrategyDifferentialEvolutionMTSv1', 'dynNpMSDEMTSv1']

    @staticmethod
    def info():
        r"""Get basic information about the algorithm.

        Returns:
            str: Basic information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""TODO"""

    def __init__(self, *args, **kwargs):
        """Initialize DynNpMultiStrategyDifferentialEvolutionMTSv1."""
        super().__init__(local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), *args, **kwargs)

    def set_parameters(self, **kwargs):
        r"""Set core parameters of DynNpMultiStrategyDifferentialEvolutionMTSv1 algorithm.

        See Also:
            * :func:`niapy.algorithm.modified.DynNpMultiStrategyDifferentialEvolutionMTS.set_parameters`

        """
        DynNpMultiStrategyDifferentialEvolutionMTS.set_parameters(self, local_searches=(mts_ls1v1, mts_ls2, mts_ls3v1), **kwargs)



import math

import numpy as np

__all__ = ['levy_flight']


def levy_flight(rng, alpha=0.01, beta=1.5, size=None):
    """Compute levy flight.

    Args:
        alpha (float): Scaling factor.
        beta (float): Stability parameter in range (0, 2).
        size (Optional[Union[int, Iterable[int]]]: Output size.
        rng (numpy.random.Generator): Random number generator.

    Returns:
        Union[float, numpy.ndarray]: Sample(s) from a truncated levy distribution.

    """
    sigma = (math.gamma(1 + beta) * np.sin(np.pi * beta / 2) / (math.gamma((1 + beta) / 2) * beta * 2 ** ((beta - 1) / 2))) ** (1 / beta)
    u = rng.normal(0, sigma, size)
    v = rng.normal(0, 1, size)
    sample = alpha * u / (np.abs(v) ** (1 / beta))
    return sample



import numpy as np

__all__ = ['euclidean']


def euclidean(u, v):
    """Compute the euclidean distance between two numpy arrays.

    Args:
        u (numpy.ndarray): Input array.
        v (numpy.ndarray): Input array.

    Returns:
        float: Euclidean distance between u and v.

    """
    return np.sqrt(np.sum(np.square(u - v), axis=-1))



# encoding=utf8
import numpy as np
import logging

from niapy.algorithms.algorithm import Algorithm

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['ClonalSelectionAlgorithm']


class ClonalSelectionAlgorithm(Algorithm):
    r"""Implementation of Clonal Selection Algorithm.

    Algorithm:
        Clonal selection algorithm

    Date:
        2021

    Authors:
        Andraž Peršon

    License:
        MIT

    Reference papers:
        * \L\. \N\. de Castro and F. J. Von Zuben. Learning and optimization using the clonal selection principle. IEEE Transactions on Evolutionary Computation, 6:239–251, 2002.
        * Brownlee, J. "Clever Algorithms: Nature-Inspired Programming Recipes" Revision 2. 2012. 280-286.

    Attributes:
        population_size (int): Population size.
        clone_factor (float): Clone factor.
        mutation_factor (float): Mutation factor.
        num_rand (int): Number of random antibodies to be added to the population each generation.
        bits_per_param (int): Number of bits per parameter of solution vector.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['ClonalSelectionAlgorithm', 'CLONALG']

    @staticmethod
    def info():
        r"""Get algorithms information.

        Returns:
            str: Algorithm information.

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r"""L. N. de Castro and F. J. Von Zuben. Learning and optimization using the clonal selection principle. IEEE Transactions on Evolutionary Computation, 6:239–251, 2002."""

    def __init__(self, population_size=10, clone_factor=0.1, mutation_factor=10.0, num_rand=1, bits_per_param=16, *args,
                 **kwargs):
        """Initialize ClonalSelectionAlgorithm.

        Args:
            population_size (Optional[int]): Population size.
            clone_factor (Optional[float]): Clone factor.
            mutation_factor (Optional[float]): Mutation factor.
            num_rand (Optional[int]): Number of random antibodies to be added to the population each generation.
            bits_per_param (Optional[int]): Number of bits per parameter of solution vector.

        See Also:
            :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, *args, **kwargs)
        self.clone_factor = clone_factor
        self.num_clones = int(self.population_size * self.clone_factor)
        self.mutation_factor = mutation_factor
        self.num_rand = num_rand
        self.bits_per_param = bits_per_param

    def set_parameters(self, population_size=10, clone_factor=0.1, mutation_factor=10.0, num_rand=1, bits_per_param=16,
                       **kwargs):
        r"""Set the parameters of the algorithm.

        Args:
            population_size (Optional[int]): Population size.
            clone_factor (Optional[float]): Clone factor.
            mutation_factor (Optional[float]): Mutation factor.
            num_rand (Optional[int]): Random number.
            bits_per_param (Optional[int]): Number of bits per parameter of solution vector.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, **kwargs)
        self.clone_factor = clone_factor
        self.num_clones = int(self.population_size * self.clone_factor)
        self.mutation_factor = mutation_factor
        self.num_rand = num_rand
        self.bits_per_param = bits_per_param

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm parameters.

        """
        params = super().get_parameters()
        params.update({
            'clone_factor': self.clone_factor,
            'mutation_factor': self.mutation_factor,
            'num_rand': self.num_rand,
            'bits_per_param': self.bits_per_param,
        })
        return params

    def decode(self, bitstrings, task):
        bits = np.flip(np.arange(self.bits_per_param))
        z = np.sum(bitstrings * 2 ** bits, axis=-1)
        return task.lower + task.range * z / (2 ** self.bits_per_param - 1)

    def evaluate(self, bitstrings, task):
        population = self.decode(bitstrings, task)
        fitness = np.apply_along_axis(task.eval, 1, population)
        return population, fitness

    def mutate(self, bitstring, mutation_rate):
        flip = self.random(bitstring.shape) > mutation_rate
        bitstring[flip] = np.logical_not(bitstring[flip])
        return bitstring

    def clone_and_hypermutate(self, bitstrings, population, population_fitness, task):
        clones = np.repeat(bitstrings, self.num_clones, axis=0)
        for i in range(clones.shape[0]):
            mutation_rate = np.exp(-self.mutation_factor * population_fitness[i // self.num_clones])
            clones[i] = self.mutate(clones[i], mutation_rate)

        clones_pop, clones_fitness = self.evaluate(clones, task)
        all_bitstrings = np.concatenate((bitstrings, clones), axis=0)
        all_population = np.concatenate((population, clones_pop), axis=0)
        all_fitness = np.concatenate((population_fitness, clones_fitness))
        sorted_ind = np.argsort(all_fitness)

        new_bitstrings = all_bitstrings[sorted_ind][:self.population_size]
        new_population = all_population[sorted_ind][:self.population_size]
        new_fitness = all_fitness[sorted_ind][:self.population_size]

        return new_bitstrings, new_population, new_fitness

    def random_insertion(self, bitstrings, population, population_fitness, task):
        if self.num_rand == 0:
            return bitstrings, population, population_fitness
        new_bitstrings = self.random((self.num_rand, task.dimension, self.bits_per_param)) > 0.5
        new_population, new_fitness = self.evaluate(new_bitstrings, task)

        all_bitstrings = np.concatenate((bitstrings, new_bitstrings), axis=0)
        all_population = np.concatenate((population, new_population), axis=0)
        all_fitness = np.concatenate((population_fitness, new_fitness))
        sorted_ind = np.argsort(all_fitness)

        next_bitstrings = all_bitstrings[sorted_ind][:self.population_size]
        next_population = all_population[sorted_ind][:self.population_size]
        next_fitness = all_fitness[sorted_ind][:self.population_size]

        return next_bitstrings, next_population, next_fitness

    def init_population(self, task):
        r"""Initialize the starting population.

        Parameters:
            task (Task): Optimization task

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray[float], Dict[str, Any]]:
                1. New population.
                2. New population fitness/function values.
                3. Additional arguments:
                    * bitstring (numpy.ndarray): Binary representation of the population.

        See Also:
            * :func:`niapy.algorithms.Algorithm.init_population`

        """
        bitstrings = self.random((self.population_size, task.dimension, self.bits_per_param)) > 0.5
        population, fitness = self.evaluate(bitstrings, task)
        return population, fitness, {'bitstrings': bitstrings}

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Clonal Selection Algorithm.

        Parameters:
            task (Task): Optimization task.
            population (numpy.ndarray): Current population
            population_fitness (numpy.ndarray[float]): Current population fitness/function values
            best_x (numpy.ndarray): Current best individual
            best_fitness (float): Current best individual function/fitness value
            params (Dict[str, Any]): Additional algorithm arguments

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, Dict[str, Any]]:
                1. New population
                2. New population fitness/function values
                3. New global best solution
                4. New global best fitness/objective value
                5. Additional arguments:
                    * bitstring (numpy.ndarray): Binary representation of the population.

        """
        bitstrings = params.pop('bitstrings')

        bitstrings, population, population_fitness = self.clone_and_hypermutate(bitstrings, population,
                                                                                population_fitness, task)
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)
        bitstrings, population, population_fitness = self.random_insertion(bitstrings, population, population_fitness,
                                                                           task)
        best_x, best_fitness = self.get_best(population, population_fitness, best_x, best_fitness)

        return population, population_fitness, best_x, best_fitness, {'bitstrings': bitstrings}



# encoding=utf8
import logging

import numpy as np

from niapy.algorithms.algorithm import Algorithm, Individual
from niapy.util import objects_to_array

logging.basicConfig()
logger = logging.getLogger('niapy.algorithms.basic')
logger.setLevel('INFO')

__all__ = ['CamelAlgorithm']


class Camel(Individual):
    r"""Implementation of population individual that is a camel for Camel algorithm.

    Algorithm:
        Camel algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Attributes:
        endurance (float): Camel endurance.
        supply (float): Camel supply.
        x_past (numpy.ndarray): Camel's past position.
        f_past (float): Camel's past function/fitness value.
        steps (int): Age of camel.

    See Also:
        * :class:`niapy.algorithms.Individual`

    """

    def __init__(self, endurance_init=None, supply_init=None, **kwargs):
        r"""Initialize the Camel.

        Args:
            endurance_init (Optional[float]): Starting endurance of Camel.
            supply_init (Optional[float]): Stating supply of Camel.

        See Also:
            * :func:`niapy.algorithms.Individual.__init__`

        """
        super().__init__(**kwargs)
        self.endurance = endurance_init
        self.endurance_past = endurance_init
        self.supply = supply_init
        self.supply_past = supply_init
        self.x_past = self.x
        self.f_past = self.f
        self.temperature = None
        self.steps = 0

    def next_temperature(self, min_temperature, max_temperature, rng):
        r"""Apply nextT function on Camel.

        Args:
            min_temperature (float): Minimum temperature.
            max_temperature (float): Maximum temperature.
            rng (numpy.random.Generator): Random number generator.

        """
        self.temperature = rng.uniform(min_temperature, max_temperature)

    def next_supply(self, burden_factor, max_iters):
        r"""Apply nextS on Camel.

        Args:
            burden_factor (float): Burden factor.
            max_iters (int): Number of Camel Algorithm iterations/generations.

        """
        self.supply = self.supply_past * (1 - burden_factor * self.steps / max_iters)

    def next_endurance(self, max_temperature, max_iters):
        r"""Apply function nextE on function on Camel.

        Args:
            max_iters (int): Number of Camel Algorithm iterations/generations
            max_temperature (float): Maximum temperature of environment

        """
        self.endurance = self.endurance_past * (1 - self.temperature / max_temperature) * (1 - self.steps / max_iters)

    def next_x(self, cb, endurance_init, supply_init, task, rng):
        r"""Apply function next_x on Camel.

        This method/function move this Camel to new position in search space.

        Args:
            cb (numpy.ndarray): Best Camel in population.
            endurance_init (float): Starting endurance of camel.
            supply_init (float): Starting supply of camel.
            task (Task): Optimization task.
            rng (numpy.random.Generator): Random number generator.

        """
        delta = rng.uniform(-1, 1)
        self.x = self.x_past + delta * (1 - (self.endurance / endurance_init)) * np.exp(
            1 - self.supply / supply_init) * (
                         cb - self.x_past)
        if not task.is_feasible(self.x):
            self.x = self.x_past
        else:
            self.f = task.eval(self.x)

    def next(self):
        r"""Save new position of Camel to old position."""
        self.x_past = self.x.copy()
        self.f_past = self.f
        self.endurance_past = self.endurance
        self.supply_past = self.supply
        self.steps += 1
        return self

    def refill(self, supply=None, endurance=None):
        r"""Apply this function to Camel.

        Args:
            supply (float): New value of Camel supply.
            endurance (float): New value of Camel endurance.

        """
        self.supply = supply
        self.endurance = endurance


class CamelAlgorithm(Algorithm):
    r"""Implementation of Camel traveling behavior.

    Algorithm:
        Camel algorithm

    Date:
        2018

    Authors:
        Klemen Berkovič

    License:
        MIT

    Reference URL:
        https://www.iasj.net/iasj?func=fulltext&aId=118375

    Reference paper:
        Ali, Ramzy. (2016). Novel Optimization Algorithm Inspired by Camel Traveling Behavior.
        Iraq J. Electrical and Electronic Engineering. 12. 167-177.

    Attributes:
        Name (List[str]): List of strings representing name of the algorithm.
        population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
        burden_factor (Optional[float]): Burden factor :math:`\in [0, 1]`.
        death_rate (Optional[float]): Dying rate :math:`\in [0, 1]`.
        visibility (Optional[float]): View range of camel.
        supply_init (Optional[float]): Initial supply :math:`\in (0, \infty)`.
        endurance_init (Optional[float]): Initial endurance :math:`\in (0, \infty)`.
        min_temperature (Optional[float]): Minimum temperature, must be true :math:`$T_{min} < T_{max}`.
        max_temperature (Optional[float]): Maximum temperature, must be true :math:`T_{min} < T_{max}`.

    See Also:
        * :class:`niapy.algorithms.Algorithm`

    """

    Name = ['CamelAlgorithm', 'CA']

    @staticmethod
    def info():
        r"""Get information about algorithm.

        Returns:
            str: Algorithm information

        See Also:
            * :func:`niapy.algorithms.Algorithm.info`

        """
        return r'''Ali, Ramzy. (2016). Novel Optimization Algorithm Inspired by Camel Traveling Behavior.
        Iraq J. Electrical and Electronic Engineering. 12. 167-177.'''

    def __init__(self, population_size=50, burden_factor=0.25, death_rate=0.5, visibility=0.5, supply_init=10,
                 endurance_init=10, min_temperature=-10, max_temperature=10, *args, **kwargs):
        r"""Initialize CamelAlgorithm.

        Args:
            population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
            burden_factor (Optional[float]): Burden factor :math:`\in [0, 1]`.
            death_rate (Optional[float]): Dying rate :math:`\in [0, 1]`.
            visibility (Optional[float]): View range of camel.
            supply_init (Optional[float]): Initial supply :math:`\in (0, \infty)`.
            endurance_init (Optional[float]): Initial endurance :math:`\in (0, \infty)`.
            min_temperature (Optional[float]): Minimum temperature, must be true :math:`$T_{min} < T_{max}`.
            max_temperature (Optional[float]): Maximum temperature, must be true :math:`T_{min} < T_{max}`.


        See Also:
            * :func:`niapy.algorithms.Algorithm.__init__`

        """
        super().__init__(population_size, individual_type=kwargs.pop('individual_type', Camel),
                         initialization_function=kwargs.pop('initialization_function', self.init_pop), *args, **kwargs)
        self.burden_factor = burden_factor
        self.death_rate = death_rate
        self.visibility = visibility
        self.supply_init = supply_init
        self.endurance_init = endurance_init
        self.min_temperature = min_temperature
        self.max_temperature = max_temperature

    def set_parameters(self, population_size=50, burden_factor=0.25, death_rate=0.5, visibility=0.5, supply_init=10,
                       endurance_init=10, min_temperature=-10, max_temperature=10, **kwargs):
        r"""Set the arguments of an algorithm.

        Args:
            population_size (Optional[int]): Population size :math:`\in [1, \infty)`.
            burden_factor (Optional[float]): Burden factor :math:`\in [0, 1]`.
            death_rate (Optional[float]): Dying rate :math:`\in [0, 1]`.
            visibility (Optional[float]): View range of camel.
            supply_init (Optional[float]): Initial supply :math:`\in (0, \infty)`.
            endurance_init (Optional[float]): Initial endurance :math:`\in (0, \infty)`.
            min_temperature (Optional[float]): Minimum temperature, must be true :math:`$T_{min} < T_{max}`.
            max_temperature (Optional[float]): Maximum temperature, must be true :math:`T_{min} < T_{max}`.

        See Also:
            * :func:`niapy.algorithms.Algorithm.set_parameters`

        """
        super().set_parameters(population_size=population_size, individual_type=Camel,
                               initialization_function=kwargs.pop('initialization_function', self.init_pop), **kwargs)
        self.burden_factor = burden_factor
        self.death_rate = death_rate
        self.visibility = visibility
        self.supply_init = supply_init
        self.endurance_init = endurance_init
        self.min_temperature = min_temperature
        self.max_temperature = max_temperature

    def get_parameters(self):
        r"""Get parameters of the algorithm.

        Returns:
            Dict[str, Any]: Algorithm Parameters.

        """
        d = super().get_parameters()
        d.update({
            'burden_factor': self.burden_factor,
            'death_rate': self.death_rate,
            'visibility': self.visibility,
            'supply_init': self.supply_init,
            'endurance_init': self.endurance_init,
            'min_temperature': self.min_temperature,
            'max_temperature': self.max_temperature
        })
        return d

    def init_pop(self, task, population_size, rng, individual_type, **_kwargs):
        r"""Initialize starting population.

        Args:
            task (Task): Optimization task.
            population_size (int): Number of camels in population.
            rng (numpy.random.Generator): Random number generator.
            individual_type (Type[Individual]): Individual type.

        Returns:
            Tuple[numpy.ndarray[Camel], numpy.ndarray[float]]:
                1. Initialize population of camels.
                2. Initialized populations function/fitness values.

        """
        caravan = objects_to_array(
            [individual_type(endurance_init=self.endurance_init, supply_init=self.supply_init, task=task, rng=rng,
                             e=True) for _ in range(population_size)])
        return caravan, np.asarray([c.f for c in caravan])

    def walk(self, camel, best_x, task):
        r"""Move the camel in search space.

        Args:
            camel (Camel): Camel that we want to move.
            best_x (numpy.ndarray): Global best coordinates.
            task (Task): Optimization task.

        Returns:
            Camel: Camel that moved in the search space.

        """
        camel.next_temperature(self.min_temperature, self.max_temperature, self.rng)
        camel.next_supply(self.burden_factor, task.max_iters)
        camel.next_endurance(self.max_temperature, task.max_iters)
        camel.next_x(best_x, self.endurance_init, self.supply_init, task, self.rng)
        return camel

    def oasis(self, c):
        r"""Apply oasis function to camel.

        Args:
            c (Camel): Camel to apply oasis on.

        Returns:
            Camel: Camel with applied oasis on.

        """
        if self.random() > (1 - self.visibility) and c.f < c.f_past:
            c.refill(self.supply_init, self.endurance_init)
        return c

    def life_cycle(self, camel, task):
        r"""Apply life cycle to Camel.

        Args:
            camel (Camel): Camel to apply life cycle.
            task (Task): Optimization task.

        Returns:
            Camel: Camel with life cycle applied to it.

        """
        if camel.f_past < self.death_rate * camel.f:
            return Camel(self.endurance_init, self.supply_init, rng=self.rng, task=task)
        else:
            return camel.next()

    def run_iteration(self, task, population, population_fitness, best_x, best_fitness, **params):
        r"""Core function of Camel Algorithm.

        Args:
            task (Task): Optimization task.
            population (numpy.ndarray[Camel]): Current population of Camels.
            population_fitness (numpy.ndarray[float]): Current population fitness/function values.
            best_x (numpy.ndarray): Current best Camel.
            best_fitness (float): Current best Camel fitness/function value.
            **params (Dict[str, Any]): Additional arguments.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, float, dict]:
                1. New population
                2. New population function/fitness value
                3. New global best solution
                4. New global best fitness/objective value
                5. Additional arguments

        """
        new_caravan = objects_to_array([self.walk(c, best_x, task) for c in population])
        new_caravan = objects_to_array([self.oasis(c) for c in new_caravan])
        new_caravan = objects_to_array([self.life_cycle(c, task) for c in new_caravan])
        new_caravan_fitness = np.asarray([c.f for c in new_caravan])
        best_x, best_fitness = self.get_best(new_caravan, new_caravan_fitness, best_x, best_fitness)
        return new_caravan, new_caravan_fitness, best_x, best_fitness, {}



# encoding=utf8

"""Implementations of Dixon Price function."""
import numpy as np

from niapy.problems.problem import Problem

__all__ = ['DixonPrice']


class DixonPrice(Problem):
    r"""Implementations of Dixon Price function.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Dixon Price Function**

        :math:`f(\textbf{x}) = (x_1 - 1)^2 + \sum_{i = 2}^D i (2x_i^2 - x_{i - 1})^2`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-10, 10]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:**
        :math:`f(\textbf{x}^*) = 0` at :math:`\textbf{x}^* = (2^{-\frac{2^1 - 2}{2^1}}, \cdots , 2^{-\frac{2^i - 2}{2^i}} , \cdots , 2^{-\frac{2^D - 2}{2^D}})`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = (x_1 - 1)^2 + \sum_{i = 2}^D i (2x_i^2 - x_{i - 1})^2$

        Equation:
            \begin{equation} f(\textbf{x}) = (x_1 - 1)^2 + \sum_{i = 2}^D i (2x_i^2 - x_{i - 1})^2 \end{equation}

        Domain:
            $-10 \leq x_i \leq 10$

    Reference:
        https://www.sfu.ca/~ssurjano/dixonpr.html

    """

    def __init__(self, dimension=4, lower=-10.0, upper=10.0, *args, **kwargs):
        r"""Initialize Dixon Price problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = (x_1 - 1)^2 + \sum_{i = 2}^D i (2x_i^2 - x_{i - 1})^2$'''

    def _evaluate(self, x):
        indices = np.arange(2, self.dimension)
        val = np.sum(indices * (2 * x[2:] ** 2 - x[1:self.dimension - 1]) ** 2)
        return (x[0] - 1) ** 2 + val



# encoding=utf8

"""Implementations of Schwefel's functions."""

import numpy as np
from niapy.problems.problem import Problem

__all__ = ['Schwefel', 'Schwefel221', 'Schwefel222', 'ModifiedSchwefel']


class Schwefel(Problem):
    r"""Implementation of Schwefel function.

    Date: 2018

    Author: Lucija Brezočnik

    License: MIT

    Function: **Schwefel function**

        :math:`f(\textbf{x}) = 418.9829d - \sum_{i=1}^{D} x_i \sin(\sqrt{\lvert x_i \rvert})`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-500, 500]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = 418.9829d - \sum_{i=1}^{D} x_i \sin(\sqrt{\lvert x_i \rvert})$

        Equation:
            \begin{equation} f(\textbf{x}) = 418.9829d - \sum_{i=1}^{D} x_i
            \sin(\sqrt{\lvert x_i \rvert}) \end{equation}

        Domain:
            $-500 \leq x_i \leq 500$

    Reference:
        https://www.sfu.ca/~ssurjano/schwef.html

    """

    def __init__(self, dimension=4, lower=-500.0, upper=500.0, *args, **kwargs):
        r"""Initialize Schwefel problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = 418.9829d - \sum_{i=1}^{D} x_i \sin(\sqrt{\lvert x_i \rvert})$'''

    def _evaluate(self, x):
        return 418.9829 * self.dimension - np.sum(x * np.sin(np.sqrt(np.abs(x))))


class Schwefel221(Problem):
    r"""Schwefel 2.21 function implementation.

    Date: 2018

    Author: Grega Vrbančič

    Licence: MIT

    Function: **Schwefel 2.21 function**

        :math:`f(\mathbf{x})=\max_{i=1,...,D}|x_i|`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x})=\max_{i=1,...,D} \lvert x_i \rvert$

        Equation:
            \begin{equation}f(\mathbf{x}) = \max_{i=1,...,D} \lvert x_i \rvert \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference paper:
        Jamil, M., and Yang, X. S. (2013).
        A literature survey of benchmark functions for global optimisation problems.
        International Journal of Mathematical Modelling and Numerical Optimisation,
        4(2), 150-194.

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Schwefel221 problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\mathbf{x})=\max_{i=1,...,D} \lvert x_i \rvert$'''

    def _evaluate(self, x):
        return np.amax(np.abs(x))


class Schwefel222(Problem):
    r"""Schwefel 2.22 function implementation.

    Date: 2018

    Author: Grega Vrbančič

    Licence: MIT

    Function: **Schwefel 2.22 function**

        :math:`f(\mathbf{x})=\sum_{i=1}^{D} \lvert x_i \rvert +\prod_{i=1}^{D} \lvert x_i \rvert`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (0,...,0)`

    LaTeX formats:
        Inline:
            $f(\mathbf{x})=\sum_{i=1}^{D} \lvert x_i \rvert +\prod_{i=1}^{D} \lvert x_i \rvert$

        Equation:
            \begin{equation}f(\mathbf{x}) = \sum_{i=1}^{D} \lvert x_i \rvert + \prod_{i=1}^{D} \lvert x_i \rvert \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference paper:
        Jamil, M., and Yang, X. S. (2013).
        A literature survey of benchmark functions for global optimisation problems.
        International Journal of Mathematical Modelling and Numerical Optimisation,
        4(2), 150-194.

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Schwefel222 problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\mathbf{x})=\sum_{i=1}^{D} \lvert x_i \rvert +\prod_{i=1}^{D} \lvert x_i \rvert$'''

    def _evaluate(self, x):
        return np.sum(np.abs(x)) + np.product(np.abs(x))


class ModifiedSchwefel(Problem):
    r"""Implementations of Modified Schwefel functions.

    Date: 2018

    Author: Klemen Berkovič

    License: MIT

    Function:
    **Modified Schwefel Function**

        :math:`f(\textbf{x}) = 418.9829 \cdot D - \sum_{i=1}^D h(x_i) \\ h(x) = g(x + 420.9687462275036)  \\ g(z) = \begin{cases} z \sin \left( \lvert z \rvert^{\frac{1}{2}} \right) &\quad \lvert z \rvert \leq 500 \\ \left( 500 - \mod (z, 500) \right) \sin \left( \sqrt{\lvert 500 - \mod (z, 500) \rvert} \right) - \frac{ \left( z - 500 \right)^2 }{ 10000 D }  &\quad z > 500 \\ \left( \mod (\lvert z \rvert, 500) - 500 \right) \sin \left( \sqrt{\lvert \mod (\lvert z \rvert, 500) - 500 \rvert} \right) + \frac{ \left( z - 500 \right)^2 }{ 10000 D } &\quad z < -500\end{cases}`

        **Input domain:**
        The function can be defined on any input domain but it is usually
        evaluated on the hypercube :math:`x_i ∈ [-100, 100]`, for all :math:`i = 1, 2,..., D`.

        **Global minimum:** :math:`f(x^*) = 0`, at :math:`x^* = (420.968746,...,420.968746)`

    LaTeX formats:
        Inline:
            $f(\textbf{x}) = 418.9829 \cdot D - \sum_{i=1}^D h(x_i) \\ h(x) = g(x + 420.9687462275036)  \\ g(z) = \begin{cases} z \sin \left( \lvert z \rvert^{\frac{1}{2}} \right) &\quad \lvert z \rvert \leq 500 \\ \left( 500 - \mod (z, 500) \right) \sin \left( \sqrt{\lvert 500 - \mod (z, 500) \rvert} \right) - \frac{ \left( z - 500 \right)^2 }{ 10000 D }  &\quad z > 500 \\ \left( \mod (\lvert z \rvert, 500) - 500 \right) \sin \left( \sqrt{\lvert \mod (\lvert z \rvert, 500) - 500 \rvert} \right) + \frac{ \left( z - 500 \right)^2 }{ 10000 D } &\quad z < -500\end{cases}$

        Equation:
            \begin{equation} f(\textbf{x}) = 418.9829 \cdot D - \sum_{i=1}^D h(x_i) \\ h(x) = g(x + 420.9687462275036)  \\ g(z) = \begin{cases} z \sin \left( \lvert z \rvert^{\frac{1}{2}} \right) &\quad \lvert z \rvert \leq 500 \\ \left( 500 - \mod (z, 500) \right) \sin \left( \sqrt{\lvert 500 - \mod (z, 500) \rvert} \right) - \frac{ \left( z - 500 \right)^2 }{ 10000 D }  &\quad z > 500 \\ \left( \mod (\lvert z \rvert, 500) - 500 \right) \sin \left( \sqrt{\lvert \mod (\lvert z \rvert, 500) - 500 \rvert} \right) + \frac{ \left( z - 500 \right)^2 }{ 10000 D } &\quad z < -500\end{cases} \end{equation}

        Domain:
            $-100 \leq x_i \leq 100$

    Reference:
        http://www5.zzu.edu.cn/__local/A/69/BC/D3B5DFE94CD2574B38AD7CD1D12_C802DAFE_BC0C0.pdf

    """

    def __init__(self, dimension=4, lower=-100.0, upper=100.0, *args, **kwargs):
        r"""Initialize Modified Schwefel problem..

        Args:
            dimension (Optional[int]): Dimension of the problem.
            lower (Optional[Union[float, Iterable[float]]]): Lower bounds of the problem.
            upper (Optional[Union[float, Iterable[float]]]): Upper bounds of the problem.

        See Also:
            :func:`niapy.problems.Problem.__init__`

        """
        super().__init__(dimension, lower, upper, *args, **kwargs)

    @staticmethod
    def latex_code():
        r"""Return the latex code of the problem.

        Returns:
            str: Latex code.

        """
        return r'''$f(\textbf{x}) = 418.9829 \cdot D - \sum_{i=1}^D h(x_i) \\ h(x) = g(x + 420.9687462275036)  \\ g(z) = \begin{cases} z \sin \left( \lvert z \rvert^{\frac{1}{2}} \right) &\quad \lvert z \rvert \leq 500 \\ \left( 500 - \mod (z, 500) \right) \sin \left( \sqrt{\lvert 500 - \mod (z, 500) \rvert} \right) - \frac{ \left( z - 500 \right)^2 }{ 10000 D }  &\quad z > 500 \\ \left( \mod (\lvert z \rvert, 500) - 500 \right) \sin \left( \sqrt{\lvert \mod (\lvert z \rvert, 500) - 500 \rvert} \right) + \frac{ \left( z - 500 \right)^2 }{ 10000 D } &\quad z < -500\end{cases}$'''

    def _evaluate(self, x):
        xx = x + 420.9687462275036
        conditions = [x > 500.0, x < -500.0]
        xx_mod = np.fmod(xx, 500.0)
        choices = [(500.0 - xx_mod) * np.sin(np.sqrt(np.abs(500.0 - xx_mod))) - (xx - 500.0) ** 2 / (10000 * self.dimension),
                   (xx_mod - 500.0) * np.sin(np.sqrt(np.abs(xx_mod - 500.0))) + (xx - 500.0) ** 2 / (10000 * self.dimension)]
        default = xx * np.sin(np.sqrt(np.abs(xx)))
        val = np.sum(np.select(conditions, choices, default=default))
        return 418.9829 * self.dimension - val
